import errno
import socket

from hazelcast.serialization.bits import *

SIZE_OF_FRAME_LENGTH_AND_FLAGS = INT_SIZE_IN_BYTES + SHORT_SIZE_IN_BYTES

_MESSAGE_TYPE_OFFSET = 0
_CORRELATION_ID_OFFSET = _MESSAGE_TYPE_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_BACKUP_ACKS_OFFSET = _CORRELATION_ID_OFFSET + LONG_SIZE_IN_BYTES
_PARTITION_ID_OFFSET = _CORRELATION_ID_OFFSET + LONG_SIZE_IN_BYTES
_FRAGMENTATION_ID_OFFSET = 0

_OUTBOUND_MESSAGE_MESSAGE_TYPE_OFFSET = _MESSAGE_TYPE_OFFSET + SIZE_OF_FRAME_LENGTH_AND_FLAGS
_OUTBOUND_MESSAGE_CORRELATION_ID_OFFSET = _CORRELATION_ID_OFFSET + SIZE_OF_FRAME_LENGTH_AND_FLAGS
_OUTBOUND_MESSAGE_PARTITION_ID_OFFSET = _PARTITION_ID_OFFSET + SIZE_OF_FRAME_LENGTH_AND_FLAGS

REQUEST_HEADER_SIZE = _OUTBOUND_MESSAGE_PARTITION_ID_OFFSET + INT_SIZE_IN_BYTES
RESPONSE_HEADER_SIZE = _RESPONSE_BACKUP_ACKS_OFFSET + BYTE_SIZE_IN_BYTES
EVENT_HEADER_SIZE = _PARTITION_ID_OFFSET + INT_SIZE_IN_BYTES

_DEFAULT_FLAGS = 0
_BEGIN_FRAGMENT_FLAG = 1 << 15
_END_FRAGMENT_FLAG = 1 << 14
_UNFRAGMENTED_MESSAGE_FLAGS = _BEGIN_FRAGMENT_FLAG | _END_FRAGMENT_FLAG
_IS_FINAL_FLAG = 1 << 13
_BEGIN_DATA_STRUCTURE_FLAG = 1 << 12
_END_DATA_STRUCTURE_FLAG = 1 << 11
_IS_NULL_FLAG = 1 << 10
_IS_EVENT_FLAG = 1 << 9
_IS_BACKUP_AWARE_FLAG = 1 << 8
_IS_BACKUP_EVENT_FLAG = 1 << 7


# For codecs
def create_initial_buffer(size, message_type, is_final=False):
    size += SIZE_OF_FRAME_LENGTH_AND_FLAGS
    buf = bytearray(size)
    LE_INT.pack_into(buf, 0, size)
    flags = _UNFRAGMENTED_MESSAGE_FLAGS
    if is_final:
        flags |= _IS_FINAL_FLAG
    LE_UINT16.pack_into(buf, INT_SIZE_IN_BYTES, flags)
    LE_INT.pack_into(buf, _OUTBOUND_MESSAGE_MESSAGE_TYPE_OFFSET, message_type)
    LE_INT.pack_into(buf, _OUTBOUND_MESSAGE_PARTITION_ID_OFFSET, -1)
    return buf


# For custom codecs
def create_initial_buffer_custom(size, is_begin_frame=False):
    size += SIZE_OF_FRAME_LENGTH_AND_FLAGS
    if is_begin_frame:
        # Needed for custom codecs that does not have initial frame at first
        # but requires later due to new fix sized parameters
        buf = bytearray(size)
        LE_INT.pack_into(buf, 0, size)
        LE_UINT16.pack_into(buf, INT_SIZE_IN_BYTES, _BEGIN_DATA_STRUCTURE_FLAG)
        return buf
    else:
        # also add BEGIN_FRAME_BUF
        buf = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS + size)
        buf[:SIZE_OF_FRAME_LENGTH_AND_FLAGS] = BEGIN_FRAME_BUF
        LE_INT.pack_into(buf, SIZE_OF_FRAME_LENGTH_AND_FLAGS, size)
        # no need to encode flags since buf is initialized with zeros
        return buf


class OutboundMessage(object):
    __slots__ = ("buf", "retryable")

    def __init__(self, buf, retryable):
        self.buf = buf
        self.retryable = retryable

    def set_correlation_id(self, correlation_id):
        LE_LONG.pack_into(self.buf, _OUTBOUND_MESSAGE_CORRELATION_ID_OFFSET, correlation_id)

    def get_correlation_id(self):
        return LE_LONG.unpack_from(self.buf, _OUTBOUND_MESSAGE_CORRELATION_ID_OFFSET)[0]

    def set_partition_id(self, partition_id):
        LE_INT.pack_into(self.buf, _OUTBOUND_MESSAGE_PARTITION_ID_OFFSET, partition_id)

    def copy(self):
        return OutboundMessage(bytearray(self.buf), self.retryable)

    def set_backup_aware_flag(self):
        flags = LE_UINT16.unpack_from(self.buf, INT_SIZE_IN_BYTES)[0]
        flags |= _IS_BACKUP_AWARE_FLAG
        LE_UINT16.pack_into(self.buf, INT_SIZE_IN_BYTES, flags)

    def __repr__(self):
        message_type = LE_INT.unpack_from(self.buf, _OUTBOUND_MESSAGE_MESSAGE_TYPE_OFFSET)[0]
        correlation_id = self.get_correlation_id()
        return "OutboundMessage(message_type=%s, correlation_id=%s, retryable=%s)" % (
            message_type,
            correlation_id,
            self.retryable,
        )


class Frame(object):
    __slots__ = ("buf", "flags", "next")

    def __init__(self, buf, flags):
        self.buf = buf
        self.flags = flags
        self.next = None

    def copy(self):
        frame = Frame(self.buf, self.flags)
        return frame

    def is_begin_frame(self):
        return self._is_flag_set(_BEGIN_DATA_STRUCTURE_FLAG)

    def is_end_frame(self):
        return self._is_flag_set(_END_DATA_STRUCTURE_FLAG)

    def is_null_frame(self):
        return self._is_flag_set(_IS_NULL_FLAG)

    def is_final_frame(self):
        return self._is_flag_set(_IS_FINAL_FLAG)

    def has_event_flag(self):
        return self._is_flag_set(_IS_EVENT_FLAG)

    def has_backup_event_flag(self):
        return self._is_flag_set(_IS_BACKUP_EVENT_FLAG)

    def has_unfragmented_message_flags(self):
        return self._is_flag_set(_UNFRAGMENTED_MESSAGE_FLAGS)

    def has_begin_fragment_flag(self):
        return self._is_flag_set(_BEGIN_FRAGMENT_FLAG)

    def has_end_fragment_flag(self):
        return self._is_flag_set(_END_FRAGMENT_FLAG)

    def _is_flag_set(self, flag_mask):
        i = self.flags & flag_mask
        return i == flag_mask


class InboundMessage(object):
    __slots__ = ("start_frame", "end_frame", "_next_frame")

    def __init__(self, start_frame):
        self.start_frame = start_frame
        self.end_frame = start_frame
        self._next_frame = start_frame

    def next_frame(self):
        result = self._next_frame
        if self._next_frame is not None:
            self._next_frame = self._next_frame.next
        return result

    def has_next_frame(self):
        return self._next_frame is not None

    def peek_next_frame(self):
        return self._next_frame

    def add_frame(self, frame):
        frame.next = None
        # For inbound messages, we always had the start_frame and end_frame set
        self.end_frame.next = frame
        self.end_frame = frame

    def get_message_type(self):
        return LE_INT.unpack_from(self.start_frame.buf, _MESSAGE_TYPE_OFFSET)[0]

    def get_correlation_id(self):
        return LE_LONG.unpack_from(self.start_frame.buf, _CORRELATION_ID_OFFSET)[0]

    def get_fragmentation_id(self):
        return LE_LONG.unpack_from(self.start_frame.buf, _FRAGMENTATION_ID_OFFSET)[0]

    def get_number_of_backup_acks(self):
        return LE_UINT8.unpack_from(self.start_frame.buf, _RESPONSE_BACKUP_ACKS_OFFSET)[0]

    def merge(self, fragment):
        # should be called after calling drop_fragmentation_frame() on fragment
        self.end_frame.next = fragment.start_frame
        self.end_frame = fragment.end_frame

    def drop_fragmentation_frame(self):
        self.start_frame = self.start_frame.next
        self._next_frame = self.start_frame

    def __repr__(self):
        message_type = self.get_message_type()
        correlation_id = self.get_correlation_id()
        return "InboundMessage(message_type=%s, correlation_id=%s)" % (message_type, correlation_id)


NULL_FRAME_BUF = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_INT.pack_into(NULL_FRAME_BUF, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_UINT16.pack_into(NULL_FRAME_BUF, INT_SIZE_IN_BYTES, _IS_NULL_FLAG)

# Has IS_NULL and IS_FINAL flags
NULL_FINAL_FRAME_BUF = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_INT.pack_into(NULL_FINAL_FRAME_BUF, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_UINT16.pack_into(NULL_FINAL_FRAME_BUF, INT_SIZE_IN_BYTES, _IS_NULL_FLAG | _IS_FINAL_FLAG)

BEGIN_FRAME_BUF = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_INT.pack_into(BEGIN_FRAME_BUF, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_UINT16.pack_into(BEGIN_FRAME_BUF, INT_SIZE_IN_BYTES, _BEGIN_DATA_STRUCTURE_FLAG)

END_FRAME_BUF = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_INT.pack_into(END_FRAME_BUF, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_UINT16.pack_into(END_FRAME_BUF, INT_SIZE_IN_BYTES, _END_DATA_STRUCTURE_FLAG)

# Has END_DATA_STRUCTURE and IS_FINAL flags
END_FINAL_FRAME_BUF = bytearray(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_INT.pack_into(END_FINAL_FRAME_BUF, 0, SIZE_OF_FRAME_LENGTH_AND_FLAGS)
LE_UINT16.pack_into(
    END_FINAL_FRAME_BUF, INT_SIZE_IN_BYTES, _END_DATA_STRUCTURE_FLAG | _IS_FINAL_FLAG
)


class ClientMessageBuilder(object):
    def __init__(self, message_callback):
        self._fragmented_messages = dict()
        self._message_callback = message_callback

    def on_message(self, client_message):
        if client_message.start_frame.has_unfragmented_message_flags():
            self._message_callback(client_message)
        else:
            fragmentation_frame = client_message.start_frame
            fragmentation_id = client_message.get_fragmentation_id()
            client_message.drop_fragmentation_frame()
            if fragmentation_frame.has_begin_fragment_flag():
                self._fragmented_messages[fragmentation_id] = client_message
            else:
                existing_message = self._fragmented_messages.get(fragmentation_id, None)
                if not existing_message:
                    raise socket.error(errno.EIO, "A message without the begin part is received.")

                existing_message.merge(client_message)
                if fragmentation_frame.has_end_fragment_flag():
                    self._message_callback(existing_message)
                    del self._fragmented_messages[fragmentation_id]



import logging

from hazelcast.errors import ClientOfflineError
from hazelcast.hash import hash_to_index

_logger = logging.getLogger(__name__)


class _PartitionTable(object):
    __slots__ = ("connection", "version", "partitions")

    def __init__(self, connection, version, partitions):
        self.connection = connection
        self.version = version
        self.partitions = partitions

    def __repr__(self):
        return "PartitionTable(connection=%s, version=%s)" % (self.connection, self.version)


class PartitionService(object):
    """
    Allows to retrieve information about the partition count, the partition owner
    or the partition id of a key.
    """

    __slots__ = ("_service", "_serialization_service")

    def __init__(self, internal_partition_service, serialization_service):
        self._service = internal_partition_service
        self._serialization_service = serialization_service

    def get_partition_owner(self, partition_id):
        """
        Returns the owner of the partition if it's set, ``None`` otherwise.

        Args:
            partition_id (int): The partition id.

        Returns:
            uuid.UUID: Owner of the partition
        """
        return self._service.get_partition_owner(partition_id)

    def get_partition_id(self, key):
        """
        Returns the partition id for a key data.

        Args:
            key: The given key.

        Returns:
            int: The partition id.
        """
        key_data = self._serialization_service.to_data(key)
        return self._service.get_partition_id(key_data)

    def get_partition_count(self):
        """
        Returns partition count of the connected cluster.

        If partition table is not fetched yet, this method returns ``0``.

        Returns:
            int: The partition count
        """
        return self._service.partition_count


class _InternalPartitionService(object):
    __slots__ = ("partition_count", "_client", "_partition_table")

    def __init__(self, client):
        self.partition_count = 0
        self._client = client
        self._partition_table = _PartitionTable(None, -1, {})

    def handle_partitions_view_event(self, connection, partitions, version):
        _logger.debug("Handling new partition table with version: %s", version)

        table = self._partition_table
        if not self._should_be_applied(connection, partitions, version, table):
            return

        new_partitions = self._prepare_partitions(partitions)
        new_table = _PartitionTable(connection, version, new_partitions)
        self._partition_table = new_table

    def get_partition_owner(self, partition_id):
        return self._partition_table.partitions.get(partition_id, None)

    def get_partition_id(self, key):
        if self.partition_count == 0:
            # Partition count can not be zero for the SYNC mode.
            # On the SYNC mode, we are waiting for the first connection to be established.
            # We are initializing the partition count with the value coming from the server with authentication.
            # This error is used only for ASYNC mode client.
            raise ClientOfflineError()
        return hash_to_index(key.get_partition_hash(), self.partition_count)

    def check_and_set_partition_count(self, partition_count):
        if self.partition_count == 0:
            self.partition_count = partition_count
            return True
        return self.partition_count == partition_count

    @classmethod
    def _should_be_applied(cls, connection, partitions, version, current):
        if not partitions:
            _logger.debug(
                "Partition view will not be applied since response is empty. "
                "Sending connection: %s, version: %s, current table: %s",
                connection,
                version,
                current,
            )
            return False

        if connection != current.connection:
            _logger.debug(
                "Partition view event coming from a new connection. Old: %s, new: %s",
                current.connection,
                connection,
            )
            return True

        if version <= current.version:
            _logger.debug(
                "Partition view will not be applied since response state version is older. "
                "Sending connection: %s, version: %s, current table: %s",
                connection,
                version,
                current,
            )
            return False

        return True

    @classmethod
    def _prepare_partitions(cls, partitions):
        new_partitions = {}
        for uuid, partition_list in partitions:
            for partition in partition_list:
                new_partitions[partition] = uuid
        return new_partitions


def string_partition_strategy(key):
    if key is None:
        return None
    try:
        index_of = key.index("@")
        return key[index_of + 1 :]
    except ValueError:
        return key



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import EntryListCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x0D0800
_REQUEST_MESSAGE_TYPE = 854016
# hex: 0x0D0801
_RESPONSE_MESSAGE_TYPE = 854017

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name, entries):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name)
    EntryListCodec.encode(buf, entries, DataCodec.encode, DataCodec.encode, True)
    return OutboundMessage(buf, False)



import binascii
import time
from datetime import datetime

from hazelcast import six
from hazelcast.core import HazelcastJsonValue
from hazelcast.serialization.bits import *
from hazelcast.serialization.api import StreamSerializer
from hazelcast.serialization.base import HazelcastSerializationError
from hazelcast.serialization.serialization_const import *
from hazelcast.six.moves import range, cPickle
from hazelcast.util import UUIDUtil

if not six.PY2:
    long = int


class BaseSerializer(StreamSerializer):
    def destroy(self):
        pass


# DEFAULT SERIALIZERS
class NoneSerializer(BaseSerializer):
    def read(self, inp):
        return None

    def write(self, out, obj):
        pass

    def get_type_id(self):
        return CONSTANT_TYPE_NULL


class BooleanSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_boolean()

    def write(self, out, obj):
        out.write_boolean(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_BOOLEAN


class ByteSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_byte()

    def write(self, out, obj):
        out.write_byte(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_BYTE


class CharSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_char()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_CHAR


class ShortSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_short()

    def write(self, out, obj):
        out.write_short(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_SHORT


class IntegerSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_int()

    def write(self, out, obj):
        out.write_int(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_INTEGER


class LongSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_long()

    def write(self, out, obj):
        out.write_long(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_LONG


class FloatSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_float()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_FLOAT


class DoubleSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_double()

    def write(self, out, obj):
        out.write_double(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_DOUBLE


class StringSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_string()

    def write(self, out, obj):
        out.write_string(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_STRING


class UuidSerializer(BaseSerializer):
    def read(self, inp):
        msb = inp.read_long()
        lsb = inp.read_long()
        return UUIDUtil.from_bits(msb, lsb)

    def write(self, out, obj):
        msb, lsb = UUIDUtil.to_bits(obj)
        out.write_long(msb)
        out.write_long(lsb)

    def get_type_id(self):
        return CONSTANT_TYPE_UUID


class HazelcastJsonValueSerializer(BaseSerializer):
    def read(self, inp):
        return HazelcastJsonValue(inp.read_string())

    def write(self, out, obj):
        out.write_string(obj.to_string())

    def get_type_id(self):
        return JAVASCRIPT_JSON_SERIALIZATION_TYPE


# ARRAY SERIALIZERS
class BooleanArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_boolean_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_BOOLEAN_ARRAY


class ByteArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_byte_array()

    def write(self, out, obj):
        out.write_byte_array(obj)

    def get_type_id(self):
        return CONSTANT_TYPE_BYTE_ARRAY


class CharArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_char_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_CHAR_ARRAY


class ShortArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_short_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_SHORT_ARRAY


class IntegerArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_int_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_INTEGER_ARRAY


class LongArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_long_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_LONG_ARRAY


class FloatArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_float_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_FLOAT_ARRAY


class DoubleArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_double_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_DOUBLE_ARRAY


class StringArraySerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_string_array()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return CONSTANT_TYPE_STRING_ARRAY


# EXTENSIONS
class DateTimeSerializer(BaseSerializer):
    def read(self, inp):
        long_time = inp.read_long()
        return datetime.fromtimestamp(long_time / 1000.0)

    def write(self, out, obj):
        long_time = long(time.mktime(obj.timetuple())) * 1000
        out.write_long(long_time)

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_DATE


class BigIntegerSerializer(BaseSerializer):
    def read(self, inp):
        length = inp.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        result = bytearray(length)
        if length > 0:
            inp.read_into(result, 0, length)
        if result[0] & 0x80:
            neg = bytearray()
            for c in result:
                neg.append(c ^ 0xFF)
            return -1 * int(binascii.hexlify(neg), 16) - 1
        return int(binascii.hexlify(result), 16)

    def write(self, out, obj):
        the_big_int = -obj - 1 if obj < 0 else obj
        end_index = -1 if (type(obj) == long and six.PY2) else None
        hex_str = hex(the_big_int)[2:end_index]
        if len(hex_str) % 2 == 1:
            prefix = "0"  # "f" if obj < 0 else "0"
            hex_str = prefix + hex_str
        num_array = bytearray(binascii.unhexlify(bytearray(hex_str, encoding="utf-8")))
        if obj < 0:
            neg = bytearray()
            for c in num_array:
                neg.append(c ^ 0xFF)
            num_array = neg
        out.write_byte_array(num_array)

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_BIG_INTEGER


class JavaClassSerializer(BaseSerializer):
    def read(self, inp):
        return inp.read_string()

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_CLASS


class ArraySerializer(BaseSerializer):
    def read(self, inp):
        size = inp.read_int()
        return [inp.read_object() for _ in range(size)]

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_ARRAY


class ArrayListSerializer(BaseSerializer):
    def read(self, inp):
        size = inp.read_int()
        if size > NULL_ARRAY_LENGTH:
            return [inp.read_object() for _ in range(0, size)]
        return None

    def write(self, out, obj):
        size = NULL_ARRAY_LENGTH if obj is None else len(obj)
        out.write_int(size)
        for i in range(0, size):
            out.write_object(obj[i])

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_ARRAY_LIST


class LinkedListSerializer(BaseSerializer):
    def read(self, inp):
        size = inp.read_int()
        if size > NULL_ARRAY_LENGTH:
            return [inp.read_object() for _ in range(0, size)]
        return None

    # "write(self, out, obj)" is never called so not implemented here

    def get_type_id(self):
        return JAVA_DEFAULT_TYPE_LINKED_LIST


class PythonObjectSerializer(BaseSerializer):
    def read(self, inp):
        str = inp.read_string().encode()
        return cPickle.loads(str)

    def write(self, out, obj):
        out.write_string(cPickle.dumps(obj, 0).decode("utf-8"))

    def get_type_id(self):
        return PYTHON_TYPE_PICKLE


class IdentifiedDataSerializer(BaseSerializer):
    def __init__(self, factories):
        self._factories = factories

    def write(self, out, obj):
        out.write_boolean(True)  # Always identified
        out.write_int(obj.get_factory_id())
        out.write_int(obj.get_class_id())
        obj.write_data(out)

    def read(self, inp):
        is_identified = inp.read_boolean()
        if not is_identified:
            raise HazelcastSerializationError(
                "Native clients only support IdentifiedDataSerializable!"
            )
        factory_id = inp.read_int()
        class_id = inp.read_int()

        factory = self._factories.get(factory_id, None)
        if factory is None:
            raise HazelcastSerializationError(
                "No DataSerializerFactory registered for namespace: %s" % factory_id
            )
        identified = factory.get(class_id, None)
        if identified is None:
            raise HazelcastSerializationError(
                "%s is not be able to create an instance for id: %s on factoryId: %s"
                % (factory, class_id, factory_id)
            )
        instance = identified()
        instance.read_data(inp)
        return instance

    def get_type_id(self):
        return CONSTANT_TYPE_DATA_SERIALIZABLE



from hazelcast.protocol.codec import (
    atomic_ref_compare_and_set_codec,
    atomic_ref_get_codec,
    atomic_ref_set_codec,
    atomic_ref_contains_codec,
    atomic_ref_apply_codec,
)
from hazelcast.proxy.cp import BaseCPProxy
from hazelcast.util import check_true, check_not_none


class AtomicReference(BaseCPProxy):
    """A distributed, highly available object reference with atomic operations.

    AtomicReference offers linearizability during crash failures and network
    partitions. It is CP with respect to the CAP principle. If a network
    partition occurs, it remains available on at most one side of the partition.

    The following are some considerations you need to know when you use AtomicReference:

    - AtomicReference works based on the byte-content and not on the object-reference.
      If you use the ``compare_and_set()`` method, do not change to the original
      value because its serialized content will then be different.
    - All methods returning an object return a private copy. You can modify the private
      copy, but the rest of the world is shielded from your changes. If you want these
      changes to be visible to the rest of the world, you need to write the change back
      to the AtomicReference; but be careful about introducing a data-race.
    - The in-memory format of an AtomicReference is ``binary``. The receiving side
      does not need to have the class definition available unless it needs to be
      deserialized on the other side., e.g., because a method like `alter()` is executed.
      This deserialization is done for every call that needs to have the object instead
      of the binary content, so be careful with expensive object graphs that need to be
      deserialized.
    - If you have an object with many fields or an object graph and you only need to
      calculate some information or need a subset of fields, you can use the `apply()`
      method. With the `apply()` method, the whole object does not need to be sent over
      the line; only the information that is relevant is sent.

    IAtomicReference does not offer exactly-once / effectively-once
    execution semantics. It goes with at-least-once execution semantics
    by default and can cause an API call to be committed multiple times
    in case of CP member failures. It can be tuned to offer at-most-once
    execution semantics. Please see `fail-on-indeterminate-operation-state`
    server-side setting.
    """

    def compare_and_set(self, expect, update):
        """Atomically sets the value to the given updated value
        only if the current value is equal to the expected value.

        Args:
            expect: The expected value.
            update: The new value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if successful, or ``False``
            if the actual value was not equal to the expected value.
        """
        expected_data = self._to_data(expect)
        new_data = self._to_data(update)
        codec = atomic_ref_compare_and_set_codec
        request = codec.encode_request(self._group_id, self._object_name, expected_data, new_data)
        return self._invoke(request, codec.decode_response)

    def get(self):
        """Gets the current value.

        Returns:
            hazelcast.future.Future[any]: The current value.
        """
        codec = atomic_ref_get_codec
        request = codec.encode_request(self._group_id, self._object_name)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def set(self, new_value):
        """Atomically sets the given value.

        Args:
            new_value: The new value.

        Returns:
            hazelcast.future.Future[None]:
        """
        new_value_data = self._to_data(new_value)
        codec = atomic_ref_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value_data, False)
        return self._invoke(request)

    def get_and_set(self, new_value):
        """Gets the old value and sets the new value.

        Args:
            new_value: The new value.

        Returns:
            hazelcast.future.Future[any]: The old value.
        """
        new_value_data = self._to_data(new_value)
        codec = atomic_ref_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value_data, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def is_none(self):
        """Checks if the stored reference is ``None``.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the stored reference is ``None``,
            ``False`` otherwise.
        """
        return self.contains(None)

    def clear(self):
        """Clears the current stored reference, so it becomes ``None``.

        Returns:
            hazelcast.future.Future[None]:
        """
        return self.set(None)

    def contains(self, value):
        """Checks if the reference contains the value.

        Args:
            value: The value to check (is allowed to be ``None``).

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the value is found, ``False`` otherwise.
        """
        value_data = self._to_data(value)
        codec = atomic_ref_contains_codec
        request = codec.encode_request(self._group_id, self._object_name, value_data)
        return self._invoke(request, codec.decode_response)

    def alter(self, function):
        """Alters the currently stored reference by applying a function on it.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            `com.hazelcast.core.IFunction` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored reference.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_ref_apply_codec
        # 0 means don't return the value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 0, True)
        return self._invoke(request)

    def alter_and_get(self, function):
        """Alters the currently stored reference by applying a function on it and
        gets the result.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            `com.hazelcast.core.IFunction` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored reference.

        Returns:
            hazelcast.future.Future[any]: The new value, the result of the applied function.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_ref_apply_codec
        # 2 means return the new value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 2, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def get_and_alter(self, function):
        """Alters the currently stored reference by applying a function on it on
        and gets the old value.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            `com.hazelcast.core.IFunction` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored reference.

        Returns:
            hazelcast.future.Future[any]: The old value, the value before the function is applied.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_ref_apply_codec
        # 1 means return the old value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 1, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def apply(self, function):
        """Applies a function on the value, the actual stored value will not
        change.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            `com.hazelcast.core.IFunction` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function applied on the currently stored reference.

        Returns:
            hazelcast.future.Future[any]: The result of the function application.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_ref_apply_codec
        # 2 means return the new value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 2, False)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec

# hex: 0x0D0E00
_REQUEST_MESSAGE_TYPE = 855552
# hex: 0x0D0E01
_RESPONSE_MESSAGE_TYPE = 855553

_REQUEST_REGISTRATION_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_REGISTRATION_ID_OFFSET + UUID_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(name, registration_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_REGISTRATION_ID_OFFSET, registration_id)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_boolean(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)



from hazelcast.serialization.api import IdentifiedDataSerializable

_PROJECTIONS_FACTORY_ID = -30


class Projection(object):
    """Marker base class for all projections.

    Projections allow the client to transform (strip down) each query result
    object in order to avoid redundant network traffic.
    """

    pass


class _AbstractProjection(Projection, IdentifiedDataSerializable):
    def write_data(self, object_data_output):
        raise NotImplementedError("write_data")

    def read_data(self, object_data_input):
        pass

    def get_factory_id(self):
        return _PROJECTIONS_FACTORY_ID

    def get_class_id(self):
        raise NotImplementedError("get_class_id")


def _validate_attribute_path(attribute_path):
    # type: (str) -> None
    if not attribute_path:
        raise ValueError("attribute_path must not be None or empty")

    if "[any]" in attribute_path:
        raise ValueError("attribute_path must not contain [any] operators")


class _SingleAttributeProjection(_AbstractProjection):
    def __init__(self, attribute_path):
        # type: (str) -> None
        _validate_attribute_path(attribute_path)
        self._attribute_path = attribute_path

    def write_data(self, object_data_output):
        object_data_output.write_string(self._attribute_path)

    def get_class_id(self):
        return 0


class _MultiAttributeProjection(_AbstractProjection):
    def __init__(self, attribute_paths):
        # type: (list[str]) -> None
        if not attribute_paths:
            raise ValueError("Specify at least one attribute path")

        for attribute_path in attribute_paths:
            _validate_attribute_path(attribute_path)

        self.attribute_paths = attribute_paths

    def write_data(self, object_data_output):
        object_data_output.write_string_array(self.attribute_paths)

    def get_class_id(self):
        return 1


class _IdentityProjection(_AbstractProjection):
    def write_data(self, object_data_output):
        pass

    def get_class_id(self):
        return 2


def single_attribute(attribute_path):
    # type: (str) -> Projection
    """Creates a projection that extracts the value of
    the given attribute path.

    Args:
        attribute_path (str): Path to extract the attribute from.

    Returns:
        Projection[any]: A projection that extracts the value of the given
        attribute path.
    """
    return _SingleAttributeProjection(attribute_path)


def multi_attribute(*attribute_paths):
    # type: (str) -> Projection
    """Creates a projection that extracts the values of
    one or more attribute paths.

    Args:
        *attribute_paths (str): Paths to extract the attributes from.

    Returns:
        Projection[list]: A projection that extracts the values of the given
        attribute paths.
    """
    return _MultiAttributeProjection(list(attribute_paths))


def identity():
    # type: () -> Projection
    """Creates a projection that does no transformation.

    Returns:
        Projection[hazelcast.core.MapEntry]: A projection that does no
        transformation.
    """
    return _IdentityProjection()



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.core import MemberInfo
from hazelcast.protocol.codec.custom.address_codec import AddressCodec
from hazelcast.protocol.builtin import MapCodec
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.codec.custom.member_version_codec import MemberVersionCodec
from hazelcast.protocol.codec.custom.endpoint_qualifier_codec import EndpointQualifierCodec

_UUID_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_UUID_DECODE_OFFSET = 0
_LITE_MEMBER_ENCODE_OFFSET = _UUID_ENCODE_OFFSET + UUID_SIZE_IN_BYTES
_LITE_MEMBER_DECODE_OFFSET = _UUID_DECODE_OFFSET + UUID_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _LITE_MEMBER_ENCODE_OFFSET + BOOLEAN_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class MemberInfoCodec(object):
    @staticmethod
    def encode(buf, member_info, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_uuid(initial_frame_buf, _UUID_ENCODE_OFFSET, member_info.uuid)
        FixSizedTypesCodec.encode_boolean(initial_frame_buf, _LITE_MEMBER_ENCODE_OFFSET, member_info.lite_member)
        buf.extend(initial_frame_buf)
        AddressCodec.encode(buf, member_info.address)
        MapCodec.encode(buf, member_info.attributes, StringCodec.encode, StringCodec.encode)
        MemberVersionCodec.encode(buf, member_info.version)
        MapCodec.encode(buf, member_info.address_map, EndpointQualifierCodec.encode, AddressCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        uuid = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _UUID_DECODE_OFFSET)
        lite_member = FixSizedTypesCodec.decode_boolean(initial_frame.buf, _LITE_MEMBER_DECODE_OFFSET)
        address = AddressCodec.decode(msg)
        attributes = MapCodec.decode(msg, StringCodec.decode, StringCodec.decode)
        version = MemberVersionCodec.decode(msg)
        is_address_map_exists = False
        address_map = None
        if not msg.peek_next_frame().is_end_frame():
            address_map = MapCodec.decode(msg, EndpointQualifierCodec.decode, AddressCodec.decode)
            is_address_map_exists = True
        CodecUtil.fast_forward_to_end_frame(msg)
        return MemberInfo(address, uuid, attributes, lite_member, version, is_address_map_exists, address_map)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import EntryListUUIDLongCodec

# hex: 0x1D0200
_REQUEST_MESSAGE_TYPE = 1901056
# hex: 0x1D0201
_RESPONSE_MESSAGE_TYPE = 1901057

_REQUEST_DELTA_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_GET_BEFORE_UPDATE_OFFSET = _REQUEST_DELTA_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_TARGET_REPLICA_UUID_OFFSET = _REQUEST_GET_BEFORE_UPDATE_OFFSET + BOOLEAN_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TARGET_REPLICA_UUID_OFFSET + UUID_SIZE_IN_BYTES
_RESPONSE_VALUE_OFFSET = RESPONSE_HEADER_SIZE
_RESPONSE_REPLICA_COUNT_OFFSET = _RESPONSE_VALUE_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, delta, get_before_update, replica_timestamps, target_replica_uuid):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_DELTA_OFFSET, delta)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_GET_BEFORE_UPDATE_OFFSET, get_before_update)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_TARGET_REPLICA_UUID_OFFSET, target_replica_uuid)
    StringCodec.encode(buf, name)
    EntryListUUIDLongCodec.encode(buf, replica_timestamps, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["value"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_VALUE_OFFSET)
    response["replica_count"] = FixSizedTypesCodec.decode_int(initial_frame.buf, _RESPONSE_REPLICA_COUNT_OFFSET)
    response["replica_timestamps"] = EntryListUUIDLongCodec.decode(msg)
    return response



import io
import logging
import random
import struct
import threading
import time
import uuid

from hazelcast import six, __version__
from hazelcast.config import ReconnectMode
from hazelcast.core import AddressHelper, CLIENT_TYPE, SERIALIZATION_VERSION
from hazelcast.errors import (
    AuthenticationError,
    TargetDisconnectedError,
    HazelcastClientNotActiveError,
    InvalidConfigurationError,
    ClientNotAllowedInClusterError,
    IllegalStateError,
    ClientOfflineError,
)
from hazelcast.future import ImmediateFuture, ImmediateExceptionFuture
from hazelcast.invocation import Invocation
from hazelcast.lifecycle import LifecycleState
from hazelcast.protocol.client_message import (
    SIZE_OF_FRAME_LENGTH_AND_FLAGS,
    Frame,
    InboundMessage,
    ClientMessageBuilder,
)
from hazelcast.protocol.codec import (
    client_authentication_codec,
    client_authentication_custom_codec,
    client_ping_codec,
)
from hazelcast.util import AtomicInteger, calculate_version, UNKNOWN_VERSION

_logger = logging.getLogger(__name__)

_INF = float("inf")


class _WaitStrategy(object):
    def __init__(self, initial_backoff, max_backoff, multiplier, cluster_connect_timeout, jitter):
        self._initial_backoff = initial_backoff
        self._max_backoff = max_backoff
        self._multiplier = multiplier
        self._cluster_connect_timeout = cluster_connect_timeout
        self._jitter = jitter
        self._attempt = None
        self._cluster_connect_attempt_begin = None
        self._current_backoff = None

        if cluster_connect_timeout == _INF:
            self._cluster_connect_timeout_text = "INFINITE"
        else:
            self._cluster_connect_timeout_text = "%.2fs" % self._cluster_connect_timeout

    def reset(self):
        self._attempt = 0
        self._cluster_connect_attempt_begin = time.time()
        self._current_backoff = min(self._max_backoff, self._initial_backoff)

    def sleep(self):
        self._attempt += 1
        time_passed = time.time() - self._cluster_connect_attempt_begin
        if time_passed > self._cluster_connect_timeout:
            _logger.warning(
                "Unable to get live cluster connection, cluster connect timeout (%s) is reached. "
                "Attempt %d.",
                self._cluster_connect_timeout_text,
                self._attempt,
            )
            return False

        # random between (-jitter * current_backoff, jitter * current_backoff)
        sleep_time = self._current_backoff + self._current_backoff * self._jitter * (
            2 * random.random() - 1
        )
        sleep_time = min(sleep_time, self._cluster_connect_timeout - time_passed)
        _logger.warning(
            "Unable to get live cluster connection, retry in %.2fs, attempt: %d, "
            "cluster connect timeout: %s, max backoff: %.2fs",
            sleep_time,
            self._attempt,
            self._cluster_connect_timeout_text,
            self._max_backoff,
        )
        time.sleep(sleep_time)
        self._current_backoff = min(self._current_backoff * self._multiplier, self._max_backoff)
        return True


class _AuthenticationStatus(object):
    AUTHENTICATED = 0
    CREDENTIALS_FAILED = 1
    SERIALIZATION_VERSION_MISMATCH = 2
    NOT_ALLOWED_IN_CLUSTER = 3


class ConnectionManager(object):
    """ConnectionManager is responsible for managing ``Connection`` objects."""

    def __init__(
        self,
        client,
        config,
        reactor,
        address_provider,
        lifecycle_service,
        partition_service,
        cluster_service,
        invocation_service,
        near_cache_manager,
    ):
        self.live = False
        self.active_connections = {}  # uuid to connection, must be modified under the _lock
        self.client_uuid = uuid.uuid4()

        self._client = client
        self._config = config
        self._reactor = reactor
        self._address_provider = address_provider
        self._lifecycle_service = lifecycle_service
        self._partition_service = partition_service
        self._cluster_service = cluster_service
        self._invocation_service = invocation_service
        self._near_cache_manager = near_cache_manager
        self._smart_routing_enabled = config.smart_routing
        self._wait_strategy = self._init_wait_strategy(config)
        self._reconnect_mode = config.reconnect_mode
        self._heartbeat_manager = _HeartbeatManager(
            self, self._client, config, reactor, invocation_service
        )
        self._connection_listeners = []
        self._connect_all_members_timer = None
        self._async_start = config.async_start
        self._connect_to_cluster_thread_running = False
        self._shuffle_member_list = config.shuffle_member_list
        self._lock = threading.RLock()
        self._connection_id_generator = AtomicInteger()
        self._labels = frozenset(config.labels)
        self._cluster_id = None
        self._load_balancer = None

    def add_listener(self, on_connection_opened=None, on_connection_closed=None):
        """Registers a ConnectionListener.

        If the same listener is registered multiple times, it will be notified multiple times.

        Args:
            on_connection_opened (function): Function to be called when a connection is opened. (Default value = None)
            on_connection_closed (function): Function to be called when a connection is removed. (Default value = None)
        """
        self._connection_listeners.append((on_connection_opened, on_connection_closed))

    def get_connection(self, member_uuid):
        return self.active_connections.get(member_uuid, None)

    def get_random_connection(self, should_get_data_member=False):
        if self._smart_routing_enabled:
            connection = self._get_connection_from_load_balancer(should_get_data_member)
            if connection:
                return connection

        # We should not get to this point under normal circumstances
        # for the smart client. For uni-socket client, there would be
        # a single connection in the dict. Therefore, copying the list
        # should be acceptable.
        for member_uuid, connection in list(six.iteritems(self.active_connections)):
            if should_get_data_member:
                member = self._cluster_service.get_member(member_uuid)
                if not member or member.lite_member:
                    continue

            return connection

        return None

    def start(self, load_balancer):
        if self.live:
            return

        self.live = True
        self._load_balancer = load_balancer
        self._heartbeat_manager.start()
        self._connect_to_cluster()

    def shutdown(self):
        if not self.live:
            return

        self.live = False
        if self._connect_all_members_timer:
            self._connect_all_members_timer.cancel()

        self._heartbeat_manager.shutdown()

        # Need to create copy of connection values to avoid modification errors on runtime
        for connection in list(six.itervalues(self.active_connections)):
            connection.close("Hazelcast client is shutting down", None)

        self.active_connections.clear()
        del self._connection_listeners[:]

    def connect_to_all_cluster_members(self, sync_start):
        if not self._smart_routing_enabled:
            return

        if sync_start:
            for member in self._cluster_service.get_members():
                try:
                    self._get_or_connect_to_member(member).result()
                except:
                    pass

        self._start_connect_all_members_timer()

    def on_connection_close(self, closed_connection):
        remote_uuid = closed_connection.remote_uuid
        remote_address = closed_connection.remote_address

        if not remote_address:
            _logger.debug(
                "Destroying %s, but it has no remote address, hence nothing is "
                "removed from the connection dictionary",
                closed_connection,
            )
            return

        with self._lock:
            connection = self.active_connections.get(remote_uuid, None)
            disconnected = False
            removed = False
            if connection == closed_connection:
                self.active_connections.pop(remote_uuid, None)
                removed = True
                _logger.info(
                    "Removed connection to %s:%s, connection: %s",
                    remote_address,
                    remote_uuid,
                    connection,
                )

                if not self.active_connections:
                    disconnected = True

        if disconnected:
            self._lifecycle_service.fire_lifecycle_event(LifecycleState.DISCONNECTED)
            self._trigger_cluster_reconnection()

        if removed:
            for _, on_connection_closed in self._connection_listeners:
                if on_connection_closed:
                    try:
                        on_connection_closed(closed_connection)
                    except:
                        _logger.exception("Exception in connection listener")
        else:
            _logger.debug(
                "Destroying %s, but there is no mapping for %s in the connection dictionary",
                closed_connection,
                remote_uuid,
            )

    def check_invocation_allowed(self):
        if self.active_connections:
            return

        if self._async_start or self._reconnect_mode == ReconnectMode.ASYNC:
            raise ClientOfflineError()
        else:
            raise IOError("No connection found to cluster")

    def _get_connection_from_load_balancer(self, should_get_data_member):
        load_balancer = self._load_balancer
        member = None
        if should_get_data_member:
            if load_balancer.can_get_next_data_member():
                member = load_balancer.next_data_member()
        else:
            member = load_balancer.next()

        if not member:
            return None

        return self.get_connection(member.uuid)

    def _get_or_connect_to_address(self, address):
        for connection in list(six.itervalues(self.active_connections)):
            if connection.remote_address == address:
                return ImmediateFuture(connection)

        try:
            translated = self._translate(address)
            connection = self._create_connection(translated)
            return self._authenticate(connection).continue_with(self._on_auth, connection)
        except Exception as e:
            return ImmediateExceptionFuture(e)

    def _get_or_connect_to_member(self, member):
        connection = self.active_connections.get(member.uuid, None)
        if connection:
            return ImmediateFuture(connection)

        try:
            translated = self._translate(member.address)
            connection = self._create_connection(translated)
            return self._authenticate(connection).continue_with(self._on_auth, connection)
        except Exception as e:
            return ImmediateExceptionFuture(e)

    def _create_connection(self, address):
        factory = self._reactor.connection_factory
        return factory(
            self,
            self._connection_id_generator.get_and_increment(),
            address,
            self._config,
            self._invocation_service.handle_client_message,
        )

    def _translate(self, address):
        translated = self._address_provider.translate(address)
        if not translated:
            raise ValueError(
                "Address provider %s could not translate address %s"
                % (self._address_provider.__class__.__name__, address)
            )

        return translated

    def _trigger_cluster_reconnection(self):
        if self._reconnect_mode == ReconnectMode.OFF:
            _logger.info("Reconnect mode is OFF. Shutting down the client")
            self._shutdown_client()
            return

        if self._lifecycle_service.running:
            self._start_connect_to_cluster_thread()

    def _init_wait_strategy(self, config):
        cluster_connect_timeout = config.cluster_connect_timeout
        if cluster_connect_timeout == -1:
            # If the no timeout is specified by the
            # user, or set to -1 explicitly, set
            # the timeout to infinite.
            cluster_connect_timeout = _INF

        return _WaitStrategy(
            config.retry_initial_backoff,
            config.retry_max_backoff,
            config.retry_multiplier,
            cluster_connect_timeout,
            config.retry_jitter,
        )

    def _start_connect_all_members_timer(self):
        connecting_uuids = set()

        def run():
            if not self._lifecycle_service.running:
                return

            for member in self._cluster_service.get_members():
                member_uuid = member.uuid

                if self.active_connections.get(member_uuid, None):
                    continue

                if member_uuid in connecting_uuids:
                    continue

                connecting_uuids.add(member_uuid)
                if not self._lifecycle_service.running:
                    break

                # Bind the bound_member_uuid to the value
                # in this loop iteration
                def cb(_, bound_member_uuid=member_uuid):
                    connecting_uuids.discard(bound_member_uuid)

                self._get_or_connect_to_member(member).add_done_callback(cb)

            self._connect_all_members_timer = self._reactor.add_timer(1, run)

        self._connect_all_members_timer = self._reactor.add_timer(1, run)

    def _connect_to_cluster(self):
        if self._async_start:
            self._start_connect_to_cluster_thread()
        else:
            self._sync_connect_to_cluster()

    def _start_connect_to_cluster_thread(self):
        with self._lock:
            if self._connect_to_cluster_thread_running:
                return

            self._connect_to_cluster_thread_running = True

        def run():
            try:
                while True:
                    self._sync_connect_to_cluster()
                    with self._lock:
                        if self.active_connections:
                            self._connect_to_cluster_thread_running = False
                            return
            except:
                _logger.exception("Could not connect to any cluster, shutting down the client")
                self._shutdown_client()

        t = threading.Thread(target=run, name="hazelcast_async_connection")
        t.daemon = True
        t.start()

    def _shutdown_client(self):
        try:
            self._client.shutdown()
        except:
            _logger.exception("Exception during client shutdown")

    def _sync_connect_to_cluster(self):
        tried_addresses = set()
        self._wait_strategy.reset()
        try:
            while True:
                tried_addresses_per_attempt = set()
                members = self._cluster_service.get_members()
                if self._shuffle_member_list:
                    random.shuffle(members)

                for member in members:
                    self._check_client_active()
                    tried_addresses_per_attempt.add(member.address)
                    connection = self._connect(member, self._get_or_connect_to_member)
                    if connection:
                        return

                for address in self._get_possible_addresses():
                    self._check_client_active()
                    if address in tried_addresses_per_attempt:
                        # We already tried this address on from the member list
                        continue

                    tried_addresses_per_attempt.add(address)
                    connection = self._connect(address, self._get_or_connect_to_address)
                    if connection:
                        return

                tried_addresses.update(tried_addresses_per_attempt)

                # If the address providers load no addresses (which seems to be possible),
                # then the above loop is not entered and the lifecycle check is missing,
                # hence we need to repeat the same check at this point.
                if not tried_addresses_per_attempt:
                    self._check_client_active()

                if not self._wait_strategy.sleep():
                    break
        except (ClientNotAllowedInClusterError, InvalidConfigurationError):
            cluster_name = self._config.cluster_name
            _logger.exception("Stopped trying on cluster %s", cluster_name)

        cluster_name = self._config.cluster_name
        _logger.info(
            "Unable to connect to any address from the cluster with name: %s. "
            "The following addresses were tried: %s",
            cluster_name,
            tried_addresses,
        )
        if self._lifecycle_service.running:
            msg = "Unable to connect to any cluster"
        else:
            msg = "Client is being shutdown"
        raise IllegalStateError(msg)

    def _connect(self, target, get_or_connect_func):
        _logger.info("Trying to connect to %s", target)
        try:
            return get_or_connect_func(target).result()
        except (ClientNotAllowedInClusterError, InvalidConfigurationError) as e:
            _logger.warning("Error during initial connection to %s", target, exc_info=True)
            raise e
        except:
            _logger.warning("Error during initial connection to %s", target, exc_info=True)
            return None

    def _authenticate(self, connection):
        client = self._client
        cluster_name = self._config.cluster_name
        client_name = client.name
        if self._config.token_provider:
            token = self._config.token_provider.token(connection.connected_address)
            request = client_authentication_custom_codec.encode_request(
                cluster_name,
                token,
                self.client_uuid,
                CLIENT_TYPE,
                SERIALIZATION_VERSION,
                __version__,
                client_name,
                self._labels,
            )
        else:
            request = client_authentication_codec.encode_request(
                cluster_name,
                self._config.creds_username,
                self._config.creds_password,
                self.client_uuid,
                CLIENT_TYPE,
                SERIALIZATION_VERSION,
                __version__,
                client_name,
                self._labels,
            )
        invocation = Invocation(
            request, connection=connection, urgent=True, response_handler=lambda m: m
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _on_auth(self, response, connection):
        try:
            response = client_authentication_codec.decode_response(response.result())
        except Exception as err:
            connection.close("Failed to authenticate connection", err)
            raise err

        status = response["status"]
        if status == _AuthenticationStatus.AUTHENTICATED:
            return self._handle_successful_auth(response, connection)

        if status == _AuthenticationStatus.CREDENTIALS_FAILED:
            err = AuthenticationError("Authentication failed. Check cluster name and credentials.")
        elif status == _AuthenticationStatus.NOT_ALLOWED_IN_CLUSTER:
            err = ClientNotAllowedInClusterError("Client is not allowed in the cluster")
        elif status == _AuthenticationStatus.SERIALIZATION_VERSION_MISMATCH:
            err = IllegalStateError("Server serialization version does not match to client")
        else:
            err = AuthenticationError(
                "Authentication status code not supported. status: %s" % status
            )

        connection.close("Failed to authenticate connection", err)
        raise err

    def _handle_successful_auth(self, response, connection):
        with self._lock:
            self._check_partition_count(response["partition_count"])

            server_version_str = response["server_hazelcast_version"]
            remote_address = response["address"]
            remote_uuid = response["member_uuid"]

            connection.remote_address = remote_address
            connection.server_version = calculate_version(server_version_str)
            connection.remote_uuid = remote_uuid

            existing = self.active_connections.get(remote_uuid, None)
            if existing:
                connection.close(
                    "Duplicate connection to same member with UUID: %s" % remote_uuid, None
                )
                return existing

            new_cluster_id = response["cluster_id"]
            changed_cluster = self._cluster_id is not None and self._cluster_id != new_cluster_id
            if changed_cluster:
                self._check_client_state_on_cluster_change(connection)
                _logger.warning(
                    "Switching from current cluster: %s to new cluster: %s",
                    self._cluster_id,
                    new_cluster_id,
                )
                self._on_cluster_restart()

            is_initial_connection = not self.active_connections
            self.active_connections[remote_uuid] = connection
            if is_initial_connection:
                self._cluster_id = new_cluster_id

        if is_initial_connection:
            self._lifecycle_service.fire_lifecycle_event(LifecycleState.CONNECTED)

        _logger.info(
            "Authenticated with server %s:%s, server version: %s, local address: %s",
            remote_address,
            remote_uuid,
            server_version_str,
            connection.local_address,
        )

        for on_connection_opened, _ in self._connection_listeners:
            if on_connection_opened:
                try:
                    on_connection_opened(connection)
                except:
                    _logger.exception("Exception in connection listener")

        if not connection.live:
            self.on_connection_close(connection)

        return connection

    def _check_client_state_on_cluster_change(self, connection):
        if self.active_connections:
            # If there are other connections, we must be connected to the wrong cluster.
            # We should not stay connected to this new connection.
            # Note that, in some racy scenarios, we might close a connection that
            # we can operate on. In those scenarios, we rely on the fact that we will
            # reopen the connections.
            reason = "Connection does not belong to the cluster %s" % self._cluster_id
            connection.close(reason, None)
            raise ValueError(reason)

    def _on_cluster_restart(self):
        self._near_cache_manager.clear_near_caches()
        self._cluster_service.clear_member_list()

    def _check_partition_count(self, partition_count):
        if not self._partition_service.check_and_set_partition_count(partition_count):
            raise ClientNotAllowedInClusterError(
                "Client can not work with this cluster because it has a "
                "different partition count. Expected partition count: %d, "
                "Member partition count: %d"
                % (self._partition_service.partition_count, partition_count)
            )

    def _check_client_active(self):
        if not self._lifecycle_service.running:
            raise HazelcastClientNotActiveError()

    def _get_possible_addresses(self):
        primaries, secondaries = self._address_provider.load_addresses()
        if self._shuffle_member_list:
            # The relative order between primary and secondary addresses should
            # not be changed. So we shuffle the lists separately and then add
            # them to the final list so that secondary addresses are not tried
            # before all primary addresses have been tried. Otherwise we can get
            # startup delays
            random.shuffle(primaries)
            random.shuffle(secondaries)

        addresses = []
        addresses.extend(primaries)
        addresses.extend(secondaries)
        return addresses


class _HeartbeatManager(object):
    _heartbeat_timer = None

    def __init__(self, connection_manager, client, config, reactor, invocation_service):
        self._connection_manager = connection_manager
        self._client = client
        self._reactor = reactor
        self._invocation_service = invocation_service
        self._heartbeat_timeout = config.heartbeat_timeout
        self._heartbeat_interval = config.heartbeat_interval

    def start(self):
        """Starts sending periodic HeartBeat operations."""

        def _heartbeat():
            conn_manager = self._connection_manager
            if not conn_manager.live:
                return

            now = time.time()
            for connection in list(six.itervalues(conn_manager.active_connections)):
                self._check_connection(now, connection)
            self._heartbeat_timer = self._reactor.add_timer(self._heartbeat_interval, _heartbeat)

        self._heartbeat_timer = self._reactor.add_timer(self._heartbeat_interval, _heartbeat)

    def shutdown(self):
        """Stops HeartBeat operations."""
        if self._heartbeat_timer:
            self._heartbeat_timer.cancel()

    def _check_connection(self, now, connection):
        if not connection.live:
            return

        if (now - connection.last_read_time) > self._heartbeat_timeout:
            _logger.warning("Heartbeat failed over the connection: %s", connection)
            connection.close(
                "Heartbeat timed out",
                TargetDisconnectedError("Heartbeat timed out to connection %s" % connection),
            )
            return

        if (now - connection.last_write_time) > self._heartbeat_interval:
            request = client_ping_codec.encode_request()
            invocation = Invocation(request, connection=connection, urgent=True)
            self._invocation_service.invoke(invocation)


_frame_header = struct.Struct("<iH")


class _Reader(object):
    def __init__(self, builder):
        self._buf = io.BytesIO()
        self._builder = builder
        self._bytes_read = 0
        self._bytes_written = 0
        # Size of the frame excluding the header (SIZE_OF_FRAME_LENGTH_AND_FLAGS bytes)
        self._frame_size = -1
        self._frame_flags = 0
        self._message = None

    def read(self, data):
        self._buf.seek(self._bytes_written)
        self._buf.write(data)
        self._bytes_written += len(data)

    def process(self):
        message = self._read_message()
        while message:
            self._builder.on_message(message)
            message = self._read_message()

    def _read_message(self):
        while True:
            if self._read_frame():
                if self._message.end_frame.is_final_frame():
                    msg = self._message
                    self._reset()
                    return msg
            else:
                return None

    def _read_frame(self):
        if self._frame_size == -1:
            if self.length < SIZE_OF_FRAME_LENGTH_AND_FLAGS:
                # we don't have even the frame length and flags ready
                return False

            self._read_frame_size_and_flags()

        if self.length < self._frame_size:
            return False

        self._buf.seek(self._bytes_read)
        size = self._frame_size
        data = self._buf.read(size)
        self._bytes_read += size
        self._frame_size = -1
        # No need to reset flags since it will be overwritten on the next read_frame_size_and_flags call
        frame = Frame(data, self._frame_flags)
        if not self._message:
            self._message = InboundMessage(frame)
        else:
            self._message.add_frame(frame)
        return True

    def _read_frame_size_and_flags(self):
        self._buf.seek(self._bytes_read)
        header_data = self._buf.read(SIZE_OF_FRAME_LENGTH_AND_FLAGS)
        self._frame_size, self._frame_flags = _frame_header.unpack_from(header_data, 0)
        self._frame_size -= SIZE_OF_FRAME_LENGTH_AND_FLAGS
        self._bytes_read += SIZE_OF_FRAME_LENGTH_AND_FLAGS

    def _reset(self):
        if self._bytes_written == self._bytes_read:
            self._buf.seek(0)
            self._buf.truncate()
            self._bytes_written = 0
            self._bytes_read = 0
        self._message = None

    @property
    def length(self):
        return self._bytes_written - self._bytes_read


class Connection(object):
    """Connection object which stores connection related information and operations."""

    def __init__(self, connection_manager, connection_id, message_callback):
        self.remote_address = None
        self.remote_uuid = None
        self.connected_address = None
        self.local_address = None
        self.last_read_time = 0
        self.last_write_time = 0
        self.start_time = 0
        self.server_version = UNKNOWN_VERSION
        self.live = True
        self.close_reason = None

        self._connection_manager = connection_manager
        self._id = connection_id
        self._builder = ClientMessageBuilder(message_callback)
        self._reader = _Reader(self._builder)

    def send_message(self, message):
        """Sends a message to this connection.

        Args:
            message (hazelcast.protocol.client_message.OutboundMessage): Message to be sent to this connection.

        Returns:
            bool: ``True`` if the message is written to the socket, ``False`` otherwise.
        """
        if not self.live:
            return False

        self._write(message.buf)
        return True

    def close(self, reason, cause):
        """Closes the connection.

        Args:
            reason (str): The reason this connection is going to be closed. Is allowed to be None.
            cause (Exception): The exception responsible for closing this connection. Is allowed to be None.
        """
        if not self.live:
            return

        self.live = False
        self.close_reason = reason
        self._log_close(reason, cause)
        try:
            self._inner_close()
        except:
            _logger.exception("Error while closing the the connection %s", self)
        self._connection_manager.on_connection_close(self)

    def _log_close(self, reason, cause):
        msg = "%s closed. Reason: %s"
        if reason:
            r = reason
        elif cause:
            r = cause
        else:
            r = "Socket explicitly closed"

        if self._connection_manager.live:
            _logger.info(msg, self, r)
        else:
            _logger.debug(msg, self, r)

    def _inner_close(self):
        raise NotImplementedError()

    def _write(self, buf):
        raise NotImplementedError()

    def __eq__(self, other):
        return isinstance(other, Connection) and self._id == other._id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self._id


class DefaultAddressProvider(object):
    """Provides initial addresses for client to find and connect to a node.

    It also provides a no-op translator.
    """

    def __init__(self, addresses):
        self._addresses = addresses

    def load_addresses(self):
        """Returns the possible primary and secondary member addresses to connect to."""
        configured_addresses = self._addresses

        if not configured_addresses:
            configured_addresses = ["127.0.0.1"]

        primaries = []
        secondaries = []
        for address in configured_addresses:
            p, s = AddressHelper.get_possible_addresses(address)
            primaries.extend(p)
            secondaries.extend(s)

        return primaries, secondaries

    def translate(self, address):
        """No-op address translator.

        It is there to provide the same API with other address providers.
        """
        return address



import logging
import sys
import threading

from hazelcast.cluster import ClusterService, _InternalClusterService
from hazelcast.config import _Config
from hazelcast.connection import ConnectionManager, DefaultAddressProvider
from hazelcast.core import DistributedObjectInfo, DistributedObjectEvent
from hazelcast.cp import CPSubsystem, ProxySessionManager
from hazelcast.discovery import HazelcastCloudAddressProvider
from hazelcast.errors import IllegalStateError
from hazelcast.invocation import InvocationService, Invocation
from hazelcast.lifecycle import LifecycleService, LifecycleState, _InternalLifecycleService
from hazelcast.listener import ListenerService, ClusterViewListenerService
from hazelcast.near_cache import NearCacheManager
from hazelcast.partition import PartitionService, _InternalPartitionService
from hazelcast.protocol.codec import (
    client_get_distributed_objects_codec,
    client_add_distributed_object_listener_codec,
    client_remove_distributed_object_listener_codec,
)
from hazelcast.proxy import (
    ProxyManager,
    MAP_SERVICE,
    QUEUE_SERVICE,
    LIST_SERVICE,
    SET_SERVICE,
    MULTI_MAP_SERVICE,
    REPLICATED_MAP_SERVICE,
    RINGBUFFER_SERVICE,
    TOPIC_SERVICE,
    RELIABLE_TOPIC_SERVICE,
    EXECUTOR_SERVICE,
    PN_COUNTER_SERVICE,
    FLAKE_ID_GENERATOR_SERVICE,
)
from hazelcast.reactor import AsyncoreReactor
from hazelcast.serialization import SerializationServiceV1
from hazelcast.sql import _InternalSqlService, SqlService
from hazelcast.statistics import Statistics
from hazelcast.transaction import TWO_PHASE, TransactionManager
from hazelcast.util import AtomicInteger, RoundRobinLB

__all__ = ("HazelcastClient",)

_logger = logging.getLogger(__name__)


class HazelcastClient(object):
    """Hazelcast client instance to access and manipulate
    distributed data structures on the Hazelcast clusters.

    Keyword Args:
        cluster_members (`list[str]`): Candidate address list that client
            will use to establish initial connection.
            By default, set to ``["127.0.0.1"]``.
        cluster_name (str): Name of the cluster to connect to. The name is
            sent as part of the the client authentication message and may
            be verified on the member. By default, set to ``dev``.
        client_name (str): Name of the client instance. By default,
            set to ``hz.client_${CLIENT_ID}``, where ``CLIENT_ID`` starts
            from ``0`` and it is incremented by ``1`` for each new client.
        connection_timeout (float): Socket timeout value in seconds for
            client to connect member nodes. Setting this to ``0`` makes
            the connection blocking. By default, set to ``5.0``.
        socket_options (`list[tuple]`): List of socket option tuples. The
            tuples must contain the parameters passed into the
            :func:`socket.setsockopt` in the same order.
        redo_operation (bool): When set to ``True``, the client will redo
            the operations that were executing on the server in case if the
            client lost connection. This can happen because of network problems,
            or simply because the member died. However it is not clear whether
            the operation was performed or not. For idempotent operations this is
            harmless, but for non idempotent ones retrying can cause to undesirable
            effects. Note that the redo can be processed on any member. By default,
            set to ``False``.
        smart_routing (bool): Enables smart mode for the client instead of
            unisocket client. Smart clients send key based operations to owner of
            the keys. Unisocket clients send all operations to a single node. By
            default, set to ``True``.
        ssl_enabled (bool): If it is ``True``, SSL is enabled. By default, set
            to ``False``.
        ssl_cafile (str): Absolute path of concatenated CA certificates used to
            validate server's certificates in PEM format. When SSL is enabled and
            cafile is not set, a set of default CA certificates from default
            locations will be used.
        ssl_certfile (str): Absolute path of the client certificate in PEM format.
        ssl_keyfile (str): Absolute path of the private key file for the client
            certificate in the PEM format. If this parameter is ``None``, private
            key will be taken from the certfile.
        ssl_password (str|bytes|bytearray|function): Password for decrypting the
            keyfile if it is encrypted. The password may be a function to call to
            get the password. It will be called with no arguments, and it should
            return a string, bytes, or bytearray. If the return value is a string
            it will be encoded as UTF-8 before using it to decrypt the key.
            Alternatively a string, bytes, or bytearray value may be supplied
            directly as the password.
        ssl_protocol (int|str): Protocol version used in SSL communication.
            By default, set to ``TLSv1_2``. See the
            :class:`hazelcast.config.SSLProtocol` for possible values.
        ssl_ciphers (str): String in the OpenSSL cipher list format to set the
            available ciphers for sockets. More than one cipher can be set by
            separating them with a colon.
        cloud_discovery_token (str): Discovery token of the Hazelcast Cloud cluster.
            When this value is set, Hazelcast Cloud discovery is enabled.
        async_start (bool): Enables non-blocking start mode of :class:`HazelcastClient`.
            When set to ``True``, the client creation will not wait to connect to
            cluster. The client instance will throw exceptions until it connects to
            cluster and becomes ready. If set to ``False``, :class:`HazelcastClient`
            will block until a cluster connection established and it is ready to use
            the client instance. By default, set to ``False``.
        reconnect_mode (int|str): Defines how a client reconnects to cluster after a
            disconnect. By default, set to ``ON``. See the
            :class:`hazelcast.config.ReconnectMode` for possible values.
        retry_initial_backoff (float): Wait period in seconds after the first failure
            before retrying. Must be non-negative. By default, set to ``1.0``.
        retry_max_backoff (float): Upper bound for the backoff interval in seconds.
            Must be non-negative. By default, set to ``30.0``.
        retry_jitter (float): Defines how much to randomize backoffs. At each iteration
            the calculated back-off is randomized via following method in pseudo-code
            ``Random(-jitter * current_backoff, jitter * current_backoff)``. Must be
            in range ``[0.0, 1.0]``. By default, set to ``0.0`` (no randomization).
        retry_multiplier (float): The factor with which to multiply backoff after a
            failed retry. Must be greater than or equal to ``1``. By default,
            set to ``1.05``.
        cluster_connect_timeout (float): Timeout value in seconds for the client to
            give up connecting to the cluster. Must be non-negative or
            equal to `-1`. By default, set to `-1`. `-1` means that the client
            will not stop trying to the target cluster. (infinite timeout)
        portable_version (int): Default value for the portable version if the
            class does not have the :func:`get_portable_version` method. Portable
            versions are used to differentiate two versions of the
            :class:`hazelcast.serialization.api.Portable` classes that have added
            or removed fields, or fields with different types.
        data_serializable_factories (dict[int, dict[int, class]]): Dictionary of
            factory id and corresponding
            :class:`hazelcast.serialization.api.IdentifiedDataSerializable` factories.
            A factory is simply a dictionary with class id and callable class
            constructors.

            .. code-block:: python

                FACTORY_ID = 1
                CLASS_ID = 1

                class SomeSerializable(IdentifiedDataSerializable):
                    # omitting the implementation
                    pass


                client = HazelcastClient(data_serializable_factories={
                    FACTORY_ID: {
                        CLASS_ID: SomeSerializable
                    }
                })

        portable_factories (dict[int, dict[int, class]]): Dictionary of
            factory id and corresponding
            :class:`hazelcast.serialization.api.Portable` factories.
            A factory is simply a dictionary with class id and callable class
            constructors.

            .. code-block:: python

                FACTORY_ID = 2
                CLASS_ID = 2

                class SomeSerializable(Portable):
                    # omitting the implementation
                    pass


                client = HazelcastClient(portable_factories={
                    FACTORY_ID: {
                        CLASS_ID: SomeSerializable
                    }
                })

        class_definitions (`list[hazelcast.serialization.portable.classdef.ClassDefinition]`):
            List of all portable class definitions.
        check_class_definition_errors (bool): When enabled, serialization system
            will check for class definitions error at start and throw an
            ``HazelcastSerializationError`` with error definition. By default,
            set to ``True``.
        is_big_endian (bool): Defines if big-endian is used as the byte order
            for the serialization. By default, set to ``True``.
        default_int_type (int|str): Defines how the ``int``/``long`` type is
            represented on the cluster side. By default, it is serialized
            as ``INT`` (``32`` bits). See the
            :class:`hazelcast.config.IntType` for possible values.
        global_serializer (hazelcast.serialization.api.StreamSerializer):
            Defines the global serializer. This serializer
            is registered as a fallback serializer to handle all other objects
            if a serializer cannot be located for them.
        custom_serializers (dict[class, hazelcast.serialization.api.StreamSerializer]):
            Dictionary of class and the corresponding custom serializers.

            .. code-block:: python

                class SomeClass(object):
                    # omitting the implementation
                    pass


                class SomeClassSerializer(StreamSerializer):
                    # omitting the implementation
                    pass

                client = HazelcastClient(custom_serializers={
                    SomeClass: SomeClassSerializer
                })

        near_caches (dict[str, dict[str, any]]): Dictionary of near cache
            names and the corresponding near cache configurations as a dictionary.
            The near cache configurations contains the following options.
            When an option is missing from the configuration, it will be
            set to its default value.

            - **invalidate_on_change** (bool): Enables cluster-assisted invalidate
              on change behavior. When set to ``True``, entries are invalidated
              when they are changed in cluster. By default, set to ``True``.
            - **in_memory_format** (int|str): Specifies in which format data will be
              stored in the Near Cache. By default, set to ``BINARY``. See the
              :class:`hazelcast.config.InMemoryFormat` for possible values.
            - **time_to_live** (float): Maximum number of seconds that an entry can
              stay in cache. When not set, entries won't be evicted due
              to expiration.
            - **max_idle** (float): Maximum number of seconds that an entry can stay
              in the Near Cache until it is accessed. When not set, entries won't be
              evicted due to inactivity.
            - **eviction_policy** (int|str): Defines eviction policy configuration.
              By default, set to ``LRU``. See the
              :class:`hazelcast.config.EvictionPolicy` for possible values.
            - **eviction_max_size** (int): Defines maximum number of entries kept in
              the memory before eviction kicks in. By default, set to ``10000``.
            - **eviction_sampling_count** (int): Number of random entries that are
              evaluated to see if some of them are already expired. By default,
              set to ``8``.
            - **eviction_sampling_pool_size** (int): Size of the pool for eviction
              candidates. The pool is kept sorted according to the eviction policy.
              By default, set to ``16``.

        load_balancer (hazelcast.util.LoadBalancer): Load balancer implementation
            for the client
        membership_listeners (`list[tuple[function, function]]`): List of membership
            listener tuples. Tuples must be of size ``2``. The first element will
            be the function to be called when a member is added, and the second
            element will be the function to be called when the member is removed
            with the :class:`hazelcast.core.MemberInfo` as the only parameter.
            Any of the elements can be ``None``, but not at the same time.
        lifecycle_listeners (`list[function]`): List of lifecycle listeners.
            Listeners will be called with the new lifecycle state as the only
            parameter when the client changes lifecycle states.
        flake_id_generators (dict[str, dict[str, any]]): Dictionary of flake id
            generator names and the corresponding flake id generator configurations
            as a dictionary. The flake id generator configurations contains the
            following options. When an option is missing from the configuration, it
            will be set to its default value.

            - **prefetch_count** (int): Defines how many IDs are pre-fetched on the
              background when a new flake id is requested from the cluster. Should
              be in the range ``1..100000``. By default, set to ``100``.
            - **prefetch_validity** (float): Defines for how long the pre-fetched IDs
              can be used. If this time elapsed, a new batch of IDs will be fetched.
              Time unit is in seconds. By default, set to ``600`` (10 minutes).

              The IDs contain timestamp component, which ensures rough global ordering
              of IDs. If an ID is assigned to an object that was created much later,
              it will be much out of order. If you don't care about ordering, set this
              value to ``0`` for unlimited ID validity.

        reliable_topics (dict[str, dict[str, any]]): Dictionary of reliable
            topic names and the corresponding reliable topic configurations as
            a dictionary. The reliable topic configurations contain the
            following options. When an option is missing from the
            configuration, it will be set to its default value.

            - **overload_policy** (int|str): Policy to handle an overloaded
              topic. By default, set to ``BLOCK``. See the
              :class:`hazelcast.config.TopicOverloadPolicy` for possible values.
            - **read_batch_size** (int): Number of messages the reliable topic
              will try to read in batch. It will get at least one, but if
              there are more available, then it will try to get more to
              increase throughput. By default, set to ``10``.

        labels (`list[str]`): Labels for the client to be sent to the cluster.
        heartbeat_interval (float): Time interval between the heartbeats sent by the
            client to the member nodes in seconds. By default, set to ``5.0``.
        heartbeat_timeout (float): If there is no message passing between the client
            and a member within the given time via this property in seconds, the
            connection will be closed. By default, set to ``60.0``.
        invocation_timeout (float): When an invocation gets an exception because

            - Member throws an exception.
            - Connection between the client and member is closed.
            - Client's heartbeat requests are timed out.

            Time passed since invocation started is compared with this property.
            If the time is already passed, then the exception is delegated to the user.
            If not, the invocation is retried. Note that, if invocation gets no exception
            and it is a long running one, then it will not get any exception, no matter
            how small this timeout is set. Time unit is in seconds. By default,
            set to ``120.0``.

        invocation_retry_pause (float): Pause time between each retry cycle of an
            invocation in seconds. By default, set to ``1.0``.
        statistics_enabled (bool): When set to ``True``, client statistics collection
            is enabled. By default, set to ``False``.
        statistics_period (float): The period in seconds the statistics run.
        shuffle_member_list (bool): Client shuffles the given member list to prevent
            all clients to connect to the same node when this property is set to
            ``True``. When it is set to ``False``, the client tries to connect to
            the nodes in the given order. By default, set to ``True``.
        backup_ack_to_client_enabled (bool): Enables client to get backup
            acknowledgements directly from the member that backups are applied,
            which reduces number of hops and increases performance for smart clients.
            This option has no effect for unisocket clients. By default, set
            to ``True`` (enabled).
        operation_backup_timeout (float): If an operation has backups, defines
            how long the invocation will wait for acks from the backup replicas
            in seconds. If acks are not received from some backups, there won't
            be any rollback on other successful replicas. By default, set to ``5.0``.
        fail_on_indeterminate_operation_state (bool): When enabled, if an operation
            has sync backups and acks are not received from backup replicas in time,
            or the member which owns primary replica of the target partition leaves
            the cluster, then the invocation fails with
            :class:`hazelcast.errors.IndeterminateOperationStateError`. However,
            even if the invocation fails, there will not be any rollback on other
            successful replicas. By default, set to ``False`` (do not fail).
        creds_username (str): Username for credentials authentication (Enterprise feature).
        creds_password (str): Password for credentials authentication (Enterprise feature).
        token_provider (hazelcast.security.TokenProvider): Token provider for
            custom authentication (Enterprise feature). Note that token_provider
            setting has priority over credentials settings.
    """

    _CLIENT_ID = AtomicInteger()

    def __init__(self, **kwargs):
        config = _Config.from_dict(kwargs)
        self._config = config
        self._context = _ClientContext()
        client_id = HazelcastClient._CLIENT_ID.get_and_increment()
        self._name = self._create_client_name(client_id)
        self._reactor = AsyncoreReactor()
        self._serialization_service = SerializationServiceV1(config)
        self._near_cache_manager = NearCacheManager(config, self._serialization_service)
        self._internal_lifecycle_service = _InternalLifecycleService(config)
        self._lifecycle_service = LifecycleService(self._internal_lifecycle_service)
        self._invocation_service = InvocationService(self, config, self._reactor)
        self._address_provider = self._create_address_provider()
        self._internal_partition_service = _InternalPartitionService(self)
        self._partition_service = PartitionService(
            self._internal_partition_service, self._serialization_service
        )
        self._internal_cluster_service = _InternalClusterService(self, config)
        self._cluster_service = ClusterService(self._internal_cluster_service)
        self._connection_manager = ConnectionManager(
            self,
            config,
            self._reactor,
            self._address_provider,
            self._internal_lifecycle_service,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._invocation_service,
            self._near_cache_manager,
        )
        self._load_balancer = self._init_load_balancer(config)
        self._listener_service = ListenerService(
            self, config, self._connection_manager, self._invocation_service
        )
        self._proxy_manager = ProxyManager(self._context)
        self._cp_subsystem = CPSubsystem(self._context)
        self._proxy_session_manager = ProxySessionManager(self._context)
        self._transaction_manager = TransactionManager(self._context)
        self._lock_reference_id_generator = AtomicInteger(1)
        self._statistics = Statistics(
            self,
            config,
            self._reactor,
            self._connection_manager,
            self._invocation_service,
            self._near_cache_manager,
        )
        self._cluster_view_listener = ClusterViewListenerService(
            self,
            self._connection_manager,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._invocation_service,
        )
        self._shutdown_lock = threading.RLock()
        self._invocation_service.init(
            self._internal_partition_service, self._connection_manager, self._listener_service
        )
        self._internal_sql_service = _InternalSqlService(
            self._connection_manager, self._serialization_service, self._invocation_service
        )
        self.sql = SqlService(self._internal_sql_service)
        self._init_context()
        self._start()

    def _init_context(self):
        self._context.init_context(
            self,
            self._config,
            self._invocation_service,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._connection_manager,
            self._serialization_service,
            self._listener_service,
            self._proxy_manager,
            self._near_cache_manager,
            self._lock_reference_id_generator,
            self._name,
            self._proxy_session_manager,
            self._reactor,
        )

    def _start(self):
        self._reactor.start()
        try:
            self._internal_lifecycle_service.start()
            self._invocation_service.start()
            membership_listeners = self._config.membership_listeners
            self._internal_cluster_service.start(self._connection_manager, membership_listeners)
            self._cluster_view_listener.start()
            self._connection_manager.start(self._load_balancer)
            sync_start = not self._config.async_start
            if sync_start:
                self._internal_cluster_service.wait_initial_member_list_fetched()
            self._connection_manager.connect_to_all_cluster_members(sync_start)

            self._listener_service.start()
            self._invocation_service.add_backup_listener()
            self._load_balancer.init(self._cluster_service)
            self._statistics.start()
        except:
            self.shutdown()
            raise
        _logger.info("Client started")

    def get_executor(self, name):
        """Creates cluster-wide ExecutorService.

        Args:
            name (str): Name of the Executor proxy.

        Returns:
            hazelcast.proxy.executor.Executor: Executor proxy for the given name.
        """
        return self._proxy_manager.get_or_create(EXECUTOR_SERVICE, name)

    def get_flake_id_generator(self, name):
        """Creates or returns a cluster-wide FlakeIdGenerator.

        Args:
            name (str): Name of the FlakeIdGenerator proxy.

        Returns:
            hazelcast.proxy.flake_id_generator.FlakeIdGenerator: FlakeIdGenerator proxy for the given name

        """
        return self._proxy_manager.get_or_create(FLAKE_ID_GENERATOR_SERVICE, name)

    def get_queue(self, name):
        """Returns the distributed queue instance with the specified name.

        Args:
            name (str): Name of the distributed queue.

        Returns:
            hazelcast.proxy.queue.Queue: Distributed queue instance with the specified name.
        """
        return self._proxy_manager.get_or_create(QUEUE_SERVICE, name)

    def get_list(self, name):
        """Returns the distributed list instance with the specified name.

        Args:
            name (str): Name of the distributed list.

        Returns:
            hazelcast.proxy.list.List: Distributed list instance with the specified name.
        """
        return self._proxy_manager.get_or_create(LIST_SERVICE, name)

    def get_map(self, name):
        """Returns the distributed map instance with the specified name.

        Args:
            name (str): Name of the distributed map.

        Returns:
            hazelcast.proxy.map.Map: Distributed map instance with the specified name.
        """
        return self._proxy_manager.get_or_create(MAP_SERVICE, name)

    def get_multi_map(self, name):
        """Returns the distributed MultiMap instance with the specified name.

        Args:
            name (str): Name of the distributed MultiMap.

        Returns:
            hazelcast.proxy.multi_map.MultiMap: Distributed MultiMap instance with the specified name.
        """
        return self._proxy_manager.get_or_create(MULTI_MAP_SERVICE, name)

    def get_pn_counter(self, name):
        """Returns the PN Counter instance with the specified name.

        Args:
            name (str): Name of the PN Counter.

        Returns:
            hazelcast.proxy.pn_counter.PNCounter: The PN Counter.
        """
        return self._proxy_manager.get_or_create(PN_COUNTER_SERVICE, name)

    def get_reliable_topic(self, name):
        """Returns the ReliableTopic instance with the specified name.

        Args:
            name (str): Name of the ReliableTopic.

        Returns:
            hazelcast.proxy.reliable_topic.ReliableTopic: The ReliableTopic.
        """
        return self._proxy_manager.get_or_create(RELIABLE_TOPIC_SERVICE, name)

    def get_replicated_map(self, name):
        """Returns the distributed ReplicatedMap instance with the specified name.

        Args:
            name (str): Name of the distributed ReplicatedMap.

        Returns:
            hazelcast.proxy.replicated_map.ReplicatedMap: Distributed ReplicatedMap instance with the specified name.
        """
        return self._proxy_manager.get_or_create(REPLICATED_MAP_SERVICE, name)

    def get_ringbuffer(self, name):
        """Returns the distributed Ringbuffer instance with the specified name.

        Args:
            name (str): Name of the distributed Ringbuffer.

        Returns:
            hazelcast.proxy.ringbuffer.Ringbuffer: Distributed RingBuffer instance with the specified name.
        """

        return self._proxy_manager.get_or_create(RINGBUFFER_SERVICE, name)

    def get_set(self, name):
        """Returns the distributed Set instance with the specified name.

        Args:
            name (str): Name of the distributed Set.

        Returns:
            hazelcast.proxy.set.Set: Distributed Set instance with the specified name.
        """
        return self._proxy_manager.get_or_create(SET_SERVICE, name)

    def get_topic(self, name):
        """Returns the Topic instance with the specified name.

        Args:
            name (str): Name of the Topic.

        Returns:
            hazelcast.proxy.topic.Topic: The Topic.
        """
        return self._proxy_manager.get_or_create(TOPIC_SERVICE, name)

    def new_transaction(self, timeout=120, durability=1, type=TWO_PHASE):
        """Creates a new Transaction associated with the current thread using default or given options.

        Args:
            timeout (int): The timeout in seconds determines the maximum lifespan of a transaction. So if a
                transaction is configured with a timeout of 2 minutes, then it will automatically rollback if it hasn't
                committed yet.
            durability (int): The durability is the number of machines that can take over if a member fails during a
                transaction commit or rollback.
            type (int): The transaction type which can be ``TWO_PHASE`` or ``ONE_PHASE``.

        Returns:
            hazelcast.transaction.Transaction: New Transaction associated with the current thread.
        """
        return self._transaction_manager.new_transaction(timeout, durability, type)

    def add_distributed_object_listener(self, listener_func):
        """Adds a listener which will be notified when a new distributed object is created or destroyed.

        Args:
            listener_func (function): Function to be called when a distributed object is created or destroyed.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        is_smart = self._config.smart_routing
        codec = client_add_distributed_object_listener_codec
        request = codec.encode_request(is_smart)

        def handle_distributed_object_event(name, service_name, event_type, source):
            event = DistributedObjectEvent(name, service_name, event_type, source)
            listener_func(event)

        def event_handler(client_message):
            return codec.handle(client_message, handle_distributed_object_event)

        return self._listener_service.register_listener(
            request,
            codec.decode_response,
            client_remove_distributed_object_listener_codec.encode_request,
            event_handler,
        )

    def remove_distributed_object_listener(self, registration_id):
        """Removes the specified distributed object listener.

        Returns silently if there is no such listener added before.

        Args:
            registration_id (str): The id of registered listener.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if registration is removed, ``False`` otherwise.
        """
        return self._listener_service.deregister_listener(registration_id)

    def get_distributed_objects(self):
        """Returns all distributed objects such as; queue, map, set, list, topic, lock, multimap.

        Also, as a side effect, it clears the local instances of the destroyed proxies.

        Returns:
            list[hazelcast.proxy.base.Proxy]: List of instances created by Hazelcast.
        """
        request = client_get_distributed_objects_codec.encode_request()
        invocation = Invocation(request, response_handler=lambda m: m)
        self._invocation_service.invoke(invocation)

        local_distributed_object_infos = {
            DistributedObjectInfo(dist_obj.service_name, dist_obj.name)
            for dist_obj in self._proxy_manager.get_distributed_objects()
        }

        response = client_get_distributed_objects_codec.decode_response(invocation.future.result())
        for dist_obj_info in response:
            local_distributed_object_infos.discard(dist_obj_info)
            self._proxy_manager.get_or_create(
                dist_obj_info.service_name, dist_obj_info.name, create_on_remote=False
            )

        for dist_obj_info in local_distributed_object_infos:
            self._proxy_manager.destroy_proxy(
                dist_obj_info.service_name, dist_obj_info.name, destroy_on_remote=False
            )

        return self._proxy_manager.get_distributed_objects()

    def shutdown(self):
        """Shuts down this HazelcastClient."""
        with self._shutdown_lock:
            if self._internal_lifecycle_service.running:
                self._internal_lifecycle_service.fire_lifecycle_event(LifecycleState.SHUTTING_DOWN)
                self._internal_lifecycle_service.shutdown()
                self._proxy_session_manager.shutdown().result()
                self._near_cache_manager.destroy_near_caches()
                self._connection_manager.shutdown()
                self._invocation_service.shutdown()
                self._statistics.shutdown()
                self._reactor.shutdown()
                self._internal_lifecycle_service.fire_lifecycle_event(LifecycleState.SHUTDOWN)

    @property
    def name(self):
        """str: Name of the client."""
        return self._name

    @property
    def lifecycle_service(self):
        """LifecycleService: Lifecycle service allows you to check if the
        client is running and add and remove lifecycle listeners.
        """
        return self._lifecycle_service

    @property
    def partition_service(self):
        """PartitionService: Partition service allows you to get partition
        count, introspect the partition owners, and partition ids of keys.
        """
        return self._partition_service

    @property
    def cluster_service(self):
        """ClusterService: Cluster service allows you to get the list of
        the cluster members and add and remove membership listeners.
        """
        return self._cluster_service

    @property
    def cp_subsystem(self):
        """CPSubsystem: CP Subsystem offers set of in-memory linearizable
        data structures.
        """
        return self._cp_subsystem

    def _create_address_provider(self):
        config = self._config
        cluster_members = config.cluster_members
        address_list_provided = len(cluster_members) > 0
        cloud_discovery_token = config.cloud_discovery_token
        cloud_enabled = cloud_discovery_token is not None
        if address_list_provided and cloud_enabled:
            raise IllegalStateError(
                "Only one discovery method can be enabled at a time. "
                "Cluster members given explicitly: %s, Hazelcast Cloud enabled: %s"
                % (address_list_provided, cloud_enabled)
            )

        if cloud_enabled:
            connection_timeout = self._get_connection_timeout(config)
            return HazelcastCloudAddressProvider(cloud_discovery_token, connection_timeout)

        return DefaultAddressProvider(cluster_members)

    def _create_client_name(self, client_id):
        client_name = self._config.client_name
        if client_name:
            return client_name
        return "hz.client_%s" % client_id

    @staticmethod
    def _get_connection_timeout(config):
        timeout = config.connection_timeout
        return sys.maxsize if timeout == 0 else timeout

    @staticmethod
    def _init_load_balancer(config):
        load_balancer = config.load_balancer
        if not load_balancer:
            load_balancer = RoundRobinLB()
        return load_balancer


class _ClientContext(object):
    """
    Context holding all the required services, managers and the configuration for a Hazelcast client.
    """

    def __init__(self):
        self.client = None
        self.config = None
        self.invocation_service = None
        self.partition_service = None
        self.cluster_service = None
        self.connection_manager = None
        self.serialization_service = None
        self.listener_service = None
        self.proxy_manager = None
        self.near_cache_manager = None
        self.lock_reference_id_generator = None
        self.name = None
        self.proxy_session_manager = None
        self.reactor = None

    def init_context(
        self,
        client,
        config,
        invocation_service,
        partition_service,
        cluster_service,
        connection_manager,
        serialization_service,
        listener_service,
        proxy_manager,
        near_cache_manager,
        lock_reference_id_generator,
        name,
        proxy_session_manager,
        reactor,
    ):
        self.client = client
        self.config = config
        self.invocation_service = invocation_service
        self.partition_service = partition_service
        self.cluster_service = cluster_service
        self.connection_manager = connection_manager
        self.serialization_service = serialization_service
        self.listener_service = listener_service
        self.proxy_manager = proxy_manager
        self.near_cache_manager = near_cache_manager
        self.lock_reference_id_generator = lock_reference_id_generator
        self.name = name
        self.proxy_session_manager = proxy_session_manager
        self.reactor = reactor



from hazelcast.protocol.codec import (
    atomic_long_add_and_get_codec,
    atomic_long_compare_and_set_codec,
    atomic_long_get_codec,
    atomic_long_get_and_add_codec,
    atomic_long_get_and_set_codec,
    atomic_long_alter_codec,
    atomic_long_apply_codec,
)
from hazelcast.proxy.cp import BaseCPProxy
from hazelcast.util import check_not_none, check_is_int


class AtomicLong(BaseCPProxy):
    """AtomicLong is a redundant and highly available distributed counter
    for 64-bit integers (``long`` type in Java).

    It works on top of the Raft consensus algorithm. It offers linearizability
    during crash failures and network partitions. It is CP with respect to
    the CAP principle. If a network partition occurs, it remains available
    on at most one side of the partition.

    AtomicLong implementation does not offer exactly-once / effectively-once
    execution semantics. It goes with at-least-once execution semantics
    by default and can cause an API call to be committed multiple times
    in case of CP member failures. It can be tuned to offer at-most-once
    execution semantics. Please see `fail-on-indeterminate-operation-state`
    server-side setting.
    """

    def add_and_get(self, delta):
        """Atomically adds the given value to the current value.

        Args:
            delta (int): The value to add to the current value.

        Returns:
            hazelcast.future.Future[int]: The updated value, the given value added
            to the current value
        """
        check_is_int(delta)
        codec = atomic_long_add_and_get_codec
        request = codec.encode_request(self._group_id, self._object_name, delta)
        return self._invoke(request, codec.decode_response)

    def compare_and_set(self, expect, update):
        """Atomically sets the value to the given updated value
        only if the current value equals the expected value.

        Args:
            expect (int): The expected value.
            update (int): The new value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if successful; or ``False`` if
            the actual value was not equal to the expected value.
        """
        check_is_int(expect)
        check_is_int(update)
        codec = atomic_long_compare_and_set_codec
        request = codec.encode_request(self._group_id, self._object_name, expect, update)
        return self._invoke(request, codec.decode_response)

    def decrement_and_get(self):
        """Atomically decrements the current value by one.

        Returns:
            hazelcast.future.Future[int]: The updated value, the current value
            decremented by one.
        """
        return self.add_and_get(-1)

    def get_and_decrement(self):
        """Atomically decrements the current value by one.

        Returns:
            hazelcast.future.Future[int]: The old value.
        """
        return self.get_and_add(-1)

    def get(self):
        """Gets the current value.

        Returns:
            hazelcast.future.Future[int]: The current value.
        """
        codec = atomic_long_get_codec
        request = codec.encode_request(self._group_id, self._object_name)
        return self._invoke(request, codec.decode_response)

    def get_and_add(self, delta):
        """Atomically adds the given value to the current value.

        Args:
            delta (int): The value to add to the current value.

        Returns:
            hazelcast.future.Future[int]: The old value before the add.
        """
        check_is_int(delta)
        codec = atomic_long_get_and_add_codec
        request = codec.encode_request(self._group_id, self._object_name, delta)
        return self._invoke(request, codec.decode_response)

    def get_and_set(self, new_value):
        """Atomically sets the given value and returns the old value.

        Args:
            new_value (int): The new value.

        Returns:
            hazelcast.future.Future[int]: The old value.
        """
        check_is_int(new_value)
        codec = atomic_long_get_and_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value)
        return self._invoke(request, codec.decode_response)

    def increment_and_get(self):
        """Atomically increments the current value by one.

        Returns:
            hazelcast.future.Future[int]: The updated value, the current value
            incremented by one.
        """
        return self.add_and_get(1)

    def get_and_increment(self):
        """Atomically increments the current value by one.

        Returns:
            hazelcast.future.Future[int]: The old value.
        """
        return self.get_and_add(1)

    def set(self, new_value):
        """Atomically sets the given value.

        Args:
            new_value (int): The new value

        Returns:
            hazelcast.future.Future[None]:
        """
        check_is_int(new_value)
        codec = atomic_long_get_and_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value)
        return self._invoke(request)

    def alter(self, function):
        """Alters the currently stored value by applying a function on it.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            ``com.hazelcast.core.IFunction`` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored value.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_long_alter_codec
        # 1 means return the new value.
        # There is no way to tell server to return nothing as of now (30.09.2020)
        # The new value is `long` (comes with the initial frame) and we
        # don't try to decode it. So, this shouldn't cause any problems.
        request = codec.encode_request(self._group_id, self._object_name, function_data, 1)
        return self._invoke(request)

    def alter_and_get(self, function):
        """Alters the currently stored value by applying a function on it and gets the result.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            ``com.hazelcast.core.IFunction`` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored value.

        Returns:
            hazelcast.future.Future[int]: The new value.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_long_alter_codec
        # 1 means return the new value.
        request = codec.encode_request(self._group_id, self._object_name, function_data, 1)
        return self._invoke(request, codec.decode_response)

    def get_and_alter(self, function):
        """Alters the currently stored value by applying a function on it on and gets the old value.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            ``com.hazelcast.core.IFunction`` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function that alters the currently stored value.

        Returns:
            hazelcast.future.Future[int]: The old value.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_long_alter_codec
        # 0 means return the old value.
        request = codec.encode_request(self._group_id, self._object_name, function_data, 0)
        return self._invoke(request, codec.decode_response)

    def apply(self, function):
        """Applies a function on the value, the actual stored value will not change.

        Notes:
            ``function`` must be an instance of ``IdentifiedDataSerializable`` or
            ``Portable`` that has a counterpart that implements the
            ``com.hazelcast.core.IFunction`` interface registered on the server-side with
            the actual implementation of the function to be applied.

        Args:
            function (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
                The function applied to the currently stored value.

        Returns:
            hazelcast.future.Future[any]: The result of the function application.
        """
        check_not_none(function, "Function cannot be None")
        function_data = self._to_data(function)
        codec = atomic_long_apply_codec
        request = codec.encode_request(self._group_id, self._object_name, function_data)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)



from hazelcast.protocol.codec import (
    transactional_multi_map_get_codec,
    transactional_multi_map_put_codec,
    transactional_multi_map_remove_codec,
    transactional_multi_map_remove_entry_codec,
    transactional_multi_map_size_codec,
    transactional_multi_map_value_count_codec,
)
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none, thread_id, ImmutableLazyDataList


class TransactionalMultiMap(TransactionalProxy):
    """Transactional implementation of :class:`~hazelcast.proxy.multi_map.MultiMap`."""

    def put(self, key, value):
        """Transactional implementation of :func:`MultiMap.put(key, value) <hazelcast.proxy.multi_map.MultiMap.put>`

        Args:
            key: The key to be stored.
            value: The value to be stored.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the size of the multimap is increased,
            ``False`` if the multimap already contains the key-value tuple.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_multi_map_put_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request, transactional_multi_map_put_codec.decode_response)

    def get(self, key):
        """Transactional implementation of :func:`MultiMap.get(key) <hazelcast.proxy.multi_map.MultiMap.get>`

        Args:
            key: The key whose associated values are returned.

        Returns:
            hazelcast.future.Future[list]: The collection of the values associated with the key.
        """
        check_not_none(key, "key can't be none")

        def handler(message):
            return ImmutableLazyDataList(
                transactional_multi_map_get_codec.decode_response(message), self._to_object
            )

        key_data = self._to_data(key)
        request = transactional_multi_map_get_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, handler)

    def remove(self, key, value):
        """Transactional implementation of :func:`MultiMap.remove(key, value)
        <hazelcast.proxy.multi_map.MultiMap.remove>`

        Args:
            key: The key of the entry to remove.
            value: The value of the entry to remove.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the item is removed, ``False`` otherwise
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = transactional_multi_map_remove_entry_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data, value_data
        )
        return self._invoke(request, transactional_multi_map_remove_entry_codec.decode_response)

    def remove_all(self, key):
        """Transactional implementation of :func:`MultiMap.remove_all(key)
        <hazelcast.proxy.multi_map.MultiMap.remove_all>`

        Args:
            key: The key of the entries to remove.

        Returns:
            hazelcast.future.Future[list]: The collection of the values associated with the key.
        """
        check_not_none(key, "key can't be none")

        def handler(message):
            return ImmutableLazyDataList(
                transactional_multi_map_remove_codec.decode_response(message), self._to_object
            )

        key_data = self._to_data(key)
        request = transactional_multi_map_remove_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, handler)

    def value_count(self, key):
        """Transactional implementation of :func:`MultiMap.value_count(key)
        <hazelcast.proxy.multi_map.MultiMap.value_count>`

        Args:
            key: The key whose number of values is to be returned.

        Returns:
            hazelcast.future.Future[int]: The number of values matching the given key in the multimap.
        """
        check_not_none(key, "key can't be none")

        key_data = self._to_data(key)
        request = transactional_multi_map_value_count_codec.encode_request(
            self.name, self.transaction.id, thread_id(), key_data
        )
        return self._invoke(request, transactional_multi_map_value_count_codec.decode_response)

    def size(self):
        """Transactional implementation of :func:`MultiMap.size() <hazelcast.proxy.multi_map.MultiMap.size>`

        Returns:
            hazelcast.future.Future[int]: the number of key-value tuples in the multimap.
        """
        request = transactional_multi_map_size_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, transactional_multi_map_size_codec.decode_response)



import logging
import sys
import threading

from hazelcast.util import AtomicInteger
from hazelcast import six

_logger = logging.getLogger(__name__)
NONE_RESULT = object()


class Future(object):
    """Future is used for representing an asynchronous computation result."""

    _result = None
    _exception = None
    _traceback = None
    _threading_locals = threading.local()

    def __init__(self):
        self._callbacks = []
        self._event = _Event()

    def set_result(self, result):
        """Sets the result of the Future.

        Args:
            result: Result of the Future.
        """
        if result is None:
            self._result = NONE_RESULT
        else:
            self._result = result
        self._event.set()
        self._invoke_callbacks()

    def set_exception(self, exception, traceback=None):
        """Sets the exception for this Future in case of errors.

        Args:
            exception (Exception): Exception to be threw in case of error.
            traceback (function): Function to be called on traceback.
        """
        if not isinstance(exception, BaseException):
            raise RuntimeError("Exception must be of BaseException type")
        self._exception = exception
        self._traceback = traceback
        self._event.set()
        self._invoke_callbacks()

    def result(self):
        """Returns the result of the Future, which makes the call synchronous if the result has not been computed yet.

        Returns:
            Result of the Future.
        """
        self._reactor_check()
        self._event.wait()
        if self._exception:
            six.reraise(self._exception.__class__, self._exception, self._traceback)
        if self._result == NONE_RESULT:
            return None
        else:
            return self._result

    def _reactor_check(self):
        if not self.done() and hasattr(self._threading_locals, "is_reactor_thread"):
            raise RuntimeError(
                "Synchronous result for incomplete operation must not be called from Reactor thread. "
                "Use add_done_callback instead."
            )

    def is_success(self):
        """Determines whether the result can be successfully computed or not."""
        return self._result is not None

    def done(self):
        """Determines whether the result is computed or not.

        Returns:
            bool: ``True`` if the result is computed, ``False`` otherwise.
        """
        return self._event.is_set()

    def running(self):
        """Determines whether the asynchronous call, the computation is still running or not.

        Returns:
            bool: ``True`` if the  result is being computed, ``False`` otherwise.
        """
        return not self.done()

    def exception(self):
        """Returns the exceptional result, if any.

        Returns:
            Exception: Exception of this Future.
        """
        self._reactor_check()
        self._event.wait()
        return self._exception

    def traceback(self):
        """Traceback function for the Future."""
        self._reactor_check()
        self._event.wait()
        return self._traceback

    def add_done_callback(self, callback):
        run_callback = False
        with self._event.condition:
            if self.done():
                run_callback = True
            else:
                self._callbacks.append(callback)

        if run_callback:
            self._invoke_cb(callback)

    def _invoke_callbacks(self):
        for callback in self._callbacks:
            self._invoke_cb(callback)

    def _invoke_cb(self, callback):
        try:
            callback(self)
        except:
            _logger.exception("Exception when invoking callback")

    def continue_with(self, continuation_func, *args):
        """Create a continuation that executes when the Future is completed.

        Args:
            continuation_func (function): A function which takes the Future as the only parameter.
                Return value of the function will be set as the result of the continuation future.
                If the return value of the function is another Future, it will be chained
                to the returned Future.
            *args: Arguments to be passed into ``continuation_function``.

        Returns:
            Future: A new Future which will be completed when the continuation is done.
        """
        future = Future()

        def callback(f):
            try:
                result = continuation_func(f, *args)
                if isinstance(result, Future):
                    future._chain(result)
                else:
                    future.set_result(result)
            except:
                exception, traceback = sys.exc_info()[1:]
                future.set_exception(exception, traceback)

        self.add_done_callback(callback)
        return future

    def _chain(self, chained_future):
        def callback(f):
            try:
                result = f.result()
                if isinstance(result, Future):
                    self._chain(result)
                else:
                    self.set_result(result)
            except:
                exception, traceback = sys.exc_info()[1:]
                self.set_exception(exception, traceback)

        chained_future.add_done_callback(callback)


class _Event(object):
    _flag = False

    def __init__(self):
        self.condition = threading.Condition(threading.Lock())

    def set(self):
        with self.condition:
            self._flag = True
            self.condition.notify_all()

    def is_set(self):
        return self._flag

    def wait(self):
        with self.condition:
            if not self._flag:
                self.condition.wait()
            return self._flag


class ImmediateFuture(Future):
    def __init__(self, result):
        self._result = result

    def set_exception(self, exception):
        raise NotImplementedError()

    def set_result(self, result):
        raise NotImplementedError()

    def done(self):
        return True

    def is_success(self):
        return True

    def exception(self):
        return None

    def traceback(self):
        return None

    def result(self):
        return self._result

    def add_done_callback(self, callback):
        self._invoke_cb(callback)


class ImmediateExceptionFuture(Future):
    def __init__(self, exception, traceback=None):
        self._exception = exception
        self._traceback = traceback

    def set_exception(self, exception, traceback=None):
        raise NotImplementedError()

    def set_result(self, result):
        raise NotImplementedError()

    def done(self):
        return True

    def is_success(self):
        return False

    def exception(self):
        return self._exception

    def traceback(self):
        return self._traceback

    def result(self):
        six.reraise(self._exception.__class__, self._exception, self._traceback)

    def add_done_callback(self, callback):
        self._invoke_cb(callback)


def combine_futures(futures):
    """Combines set of Futures.

    Args:
        futures (list[Future]): List of Futures to be combined.

    Returns:
        Future: Result of the combination.
    """
    expected = len(futures)
    results = []
    if expected == 0:
        return ImmediateFuture(results)

    completed = AtomicInteger()
    combined = Future()

    def done(f):
        if not combined.done():
            if f.is_success():  # TODO: ensure ordering of results as original list
                results.append(f.result())
                if completed.get_and_increment() + 1 == expected:
                    combined.set_result(results)
            else:
                combined.set_exception(f.exception(), f.traceback())

    for future in futures:
        future.add_done_callback(done)

    return combined


class _BlockingWrapper(object):
    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattr__(self, item):
        inner = getattr(self._wrapped, item)
        if callable(inner):
            return self.wrap(inner)
        return inner

    def wrap(self, inner):
        def f(*args, **kwargs):
            result = inner(*args, **kwargs)
            if isinstance(result, Future):
                return result.result()
            return result

        return f

    def __repr__(self):
        return self._wrapped.__repr__()


def make_blocking(instance):
    """Takes an instance and returns an object whose methods which return non-blocking Future become blocking calls.

    Args:
        instance: A non-blocking instance.

    Returns:
        Blocking version of given non-blocking instance.
    """
    return _BlockingWrapper(instance)



from hazelcast.errors import IllegalStateError
from hazelcast.protocol.codec import (
    queue_add_all_codec,
    queue_add_listener_codec,
    queue_clear_codec,
    queue_compare_and_remove_all_codec,
    queue_compare_and_retain_all_codec,
    queue_contains_all_codec,
    queue_contains_codec,
    queue_drain_to_max_size_codec,
    queue_is_empty_codec,
    queue_iterator_codec,
    queue_offer_codec,
    queue_peek_codec,
    queue_poll_codec,
    queue_put_codec,
    queue_remaining_capacity_codec,
    queue_remove_codec,
    queue_remove_listener_codec,
    queue_size_codec,
    queue_take_codec,
)
from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.util import check_not_none, to_millis, ImmutableLazyDataList


class Queue(PartitionSpecificProxy):
    """Concurrent, blocking, distributed, observable queue.

    Queue is not a partitioned data-structure. All of the Queue content is stored in
    a single machine (and in the backup). Queue will not scale by adding more members in the cluster.
    """

    def add(self, item):
        """Adds the specified item to this queue if there is available space.

        Args:
            item: The specified item.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if element is successfully added, ``False`` otherwise.
        """

        def result_fnc(f):
            if f.result():
                return True
            raise IllegalStateError("Queue is full!")

        return self.offer(item).continue_with(result_fnc)

    def add_all(self, items):
        """Adds the elements in the specified collection to this queue.

        Args:
            items (list): Collection which includes the items to be added.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this queue is changed after call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = queue_add_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_add_all_codec.decode_response)

    def add_listener(self, include_value=False, item_added_func=None, item_removed_func=None):
        """Adds an item listener for this queue. Listener will be notified for all queue add/remove events.

        Args:
            include_value (bool): Whether received events include the updated item or not.
            item_added_func (function): Function to be called when an item is added to this set.
            item_removed_func (function): Function to be called when an item is deleted from this set.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        codec = queue_add_listener_codec
        request = codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item, uuid, event_type):
            item = item if include_value else None
            member = self._context.cluster_service.get_member(uuid)

            item_event = ItemEvent(self.name, item, event_type, member, self._to_object)
            if event_type == ItemEventType.ADDED:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: queue_remove_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, handle_event_item),
        )

    def clear(self):
        """Clears this queue. Queue will be empty after this call.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = queue_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def contains(self, item):
        """Determines whether this queue contains the specified item or not.

        Args:
            item: The specified item to be searched.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified item exists in this queue, ``False`` otherwise.
        """
        check_not_none(item, "Item can't be None")
        item_data = self._to_data(item)
        request = queue_contains_codec.encode_request(self.name, item_data)
        return self._invoke(request, queue_contains_codec.decode_response)

    def contains_all(self, items):
        """Determines whether this queue contains all of the items in the specified collection or not.

        Args:
            items (list): The specified collection which includes the items to be searched.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if all of the items in the specified collection exist
            in this queue, ``False`` otherwise.
        """
        check_not_none(items, "Items can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "item can't be None")
            data_items.append(self._to_data(item))

        request = queue_contains_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_contains_all_codec.decode_response)

    def drain_to(self, target_list, max_size=-1):
        """Transfers all available items to the given `target_list` and removes these items from this queue.

        If a max_size is specified, it transfers at most the given number of items.
        In case of a failure, an item can exist in both collections or none of them.

        This operation may be more efficient than polling elements repeatedly and putting into collection.

        Args:
            target_list (list): the list where the items in this queue will be transferred.
            max_size (int): The maximum number items to transfer.

        Returns:
            hazelcast.future.Future[int]: Number of transferred items.
        """

        def handler(message):
            response = queue_drain_to_max_size_codec.decode_response(message)
            target_list.extend(map(self._to_object, response))
            return len(response)

        request = queue_drain_to_max_size_codec.encode_request(self.name, max_size)
        return self._invoke(request, handler)

    def iterator(self):
        """Returns all of the items in this queue.

        Returns:
            list: Collection of items in this queue.
        """

        def handler(message):
            return ImmutableLazyDataList(
                queue_iterator_codec.decode_response(message), self._to_object
            )

        request = queue_iterator_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def is_empty(self):
        """Determines whether this set is empty or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this queue is empty, ``False`` otherwise.
        """
        request = queue_is_empty_codec.encode_request(self.name)
        return self._invoke(request, queue_is_empty_codec.decode_response)

    def offer(self, item, timeout=0):
        """Inserts the specified element into this queue if it is possible to do so immediately
        without violating capacity restrictions.

        If there is no space currently available:

        - If the timeout is provided, it waits until this timeout elapses and returns the result.
        - If the timeout is not provided, returns ``False`` immediately.

        Args:
            item: The item to be added.
            timeout (int): Maximum time in seconds to wait for addition.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the element was added to this queue, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        request = queue_offer_codec.encode_request(self.name, element_data, to_millis(timeout))
        return self._invoke(request, queue_offer_codec.decode_response)

    def peek(self):
        """Retrieves the head of queue without removing it from the queue.

        Returns:
            hazelcast.future.Future[any]: the head of this queue, or ``None`` if this queue is empty.
        """

        def handler(message):
            return self._to_object(queue_peek_codec.decode_response(message))

        request = queue_peek_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def poll(self, timeout=0):
        """Retrieves and removes the head of this queue.

        If this queue is empty:

        - If the timeout is provided, it waits until this timeout elapses and returns the result.
        - If the timeout is not provided, returns ``None``.

        Args:
            timeout (int): Maximum time in seconds to wait for addition.

        Returns:
            hazelcast.future.Future[any]: The head of this queue, or ``None`` if this queue is empty
            or specified timeout elapses before an item is added to the queue.
        """

        def handler(message):
            return self._to_object(queue_poll_codec.decode_response(message))

        request = queue_poll_codec.encode_request(self.name, to_millis(timeout))
        return self._invoke(request, handler)

    def put(self, item):
        """Adds the specified element into this queue.

        If there is no space, it waits until necessary space becomes available.

        Args:
            item: The specified item.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        request = queue_put_codec.encode_request(self.name, element_data)
        return self._invoke(request)

    def remaining_capacity(self):
        """Returns the remaining capacity of this queue.

        Returns:
            hazelcast.future.Future[int]: Remaining capacity of this queue.
        """
        request = queue_remaining_capacity_codec.encode_request(self.name)
        return self._invoke(request, queue_remaining_capacity_codec.decode_response)

    def remove(self, item):
        """Removes the specified element from the queue if it exists.

        Args:
            item: The specified element to be removed.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified element exists in this queue, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        request = queue_remove_codec.encode_request(self.name, item_data)
        return self._invoke(request, queue_remove_codec.decode_response)

    def remove_all(self, items):
        """Removes all of the elements of the specified collection from this queue.

        Args:
            items (list): The specified collection.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the call changed this queue, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = queue_compare_and_remove_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_compare_and_remove_all_codec.decode_response)

    def remove_listener(self, registration_id):
        """Removes the specified item listener.

        Returns silently if the specified listener was not added before.

        Args:
            registration_id (str): Id of the listener to be deleted.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the item listener is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items):
        """Removes the items which are not contained in the specified collection.

        In other words, only the items that are contained in the specified collection will be retained.

        Args:
            items (list): Collection which includes the elements to be retained in this set.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this queue changed as a result of the call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        request = queue_compare_and_retain_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_compare_and_retain_all_codec.decode_response)

    def size(self):
        """Returns the number of elements in this collection.

        If the size is greater than ``2**31 - 1``, it returns ``2**31 - 1``

        Returns:
            hazelcast.future.Future[int]: Size of the queue.
        """
        request = queue_size_codec.encode_request(self.name)
        return self._invoke(request, queue_size_codec.decode_response)

    def take(self):
        """Retrieves and removes the head of this queue, if necessary, waits until an item becomes available.

        Returns:
            hazelcast.future.Future[any]: The head of this queue.
        """

        def handler(message):
            return self._to_object(queue_take_codec.decode_response(message))

        request = queue_take_codec.encode_request(self.name)
        return self._invoke(request, handler)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec

# hex: 0x1C0100
_REQUEST_MESSAGE_TYPE = 1835264
# hex: 0x1C0101
_RESPONSE_MESSAGE_TYPE = 1835265

_REQUEST_BATCH_SIZE_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_BATCH_SIZE_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_BASE_OFFSET = RESPONSE_HEADER_SIZE
_RESPONSE_INCREMENT_OFFSET = _RESPONSE_BASE_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_BATCH_SIZE_OFFSET = _RESPONSE_INCREMENT_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, batch_size):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_BATCH_SIZE_OFFSET, batch_size)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["base"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_BASE_OFFSET)
    response["increment"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_INCREMENT_OFFSET)
    response["batch_size"] = FixSizedTypesCodec.decode_int(initial_frame.buf, _RESPONSE_BATCH_SIZE_OFFSET)
    return response



from hazelcast import six
from hazelcast.hash import murmur_hash3_x86_32
from hazelcast.serialization import BE_INT
from hazelcast.serialization.serialization_const import *

PARTITION_HASH_OFFSET = 0
TYPE_OFFSET = 4
DATA_OFFSET = 8

HEAP_DATA_OVERHEAD = DATA_OFFSET


class Data(object):
    """Data is basic unit of serialization.

    It stores binary form of an object serialized by serialization service.
    """

    def __init__(self, buf):
        self._buffer = buf

    def to_bytes(self):
        """Returns byte array representation of internal binary format.

        Returns:
            bytearray: The byte array representation of internal binary format.
        """
        return self._buffer

    def get_type(self):
        """Returns serialization type of binary form.

        Returns:
            int: Serialization type of binary form.
        """
        if len(self._buffer) == 0:
            return CONSTANT_TYPE_NULL
        return BE_INT.unpack_from(self._buffer, TYPE_OFFSET)[0]

    def total_size(self):
        """Returns the total size of Data in bytes.

        Returns:
            int: Total size of Data in bytes.
        """
        return len(self._buffer)

    def data_size(self):
        """Returns size of internal binary data in bytes.

        Returns:
            int: Size of internal binary data in bytes.
        """
        return max(len(self._buffer) - HEAP_DATA_OVERHEAD, 0)

    def get_partition_hash(self):
        """Returns partition hash calculated for serialized object.

        Partition hash is used to determine partition of a Data and is calculated using:

        - PartitioningStrategy during serialization.
        - If partition hash is not set then hash_code() is used.

        Returns:
            int: Partition hash.
        """
        partition_hash = BE_INT.unpack_from(self._buffer, PARTITION_HASH_OFFSET)[0]
        if partition_hash != 0:
            return partition_hash
        return self.hash_code()

    def is_portable(self):
        """Determines whether this Data is created from a ``Portable`` object or not.

        Returns:
            bool: ``True`` if source object is Portable, ``False`` otherwise.
        """
        return CONSTANT_TYPE_PORTABLE == self.get_type()

    def hash_code(self):
        """Returns the murmur hash of the internal data.

        Returns:
            int: The murmur hash of the internal data.
        """
        return murmur_hash3_x86_32(self._buffer)

    def __hash__(self):
        # Data objects are used in NearCache as keys.
        # When this method is called on Data objects
        # received from the members, buffer is type
        # of bytes instead of bytearray. Since bytes
        # is an alias of str in Python2, we cannot
        # use murmur hash directly on it. The
        # conversion is necessary only on Python2
        if isinstance(self._buffer, bytearray) or six.PY3:
            return murmur_hash3_x86_32(self._buffer)
        return murmur_hash3_x86_32(bytearray(self._buffer))

    def __eq__(self, other):
        return (
            isinstance(other, Data)
            and self.total_size() == other.total_size()
            and self._buffer == other.to_bytes()
        )

    def __len__(self):
        return self.total_size()



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import ByteArrayCodec
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.codec.custom.address_codec import AddressCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x000200
_REQUEST_MESSAGE_TYPE = 512
# hex: 0x000201
_RESPONSE_MESSAGE_TYPE = 513

_REQUEST_UUID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_SERIALIZATION_VERSION_OFFSET = _REQUEST_UUID_OFFSET + UUID_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_SERIALIZATION_VERSION_OFFSET + BYTE_SIZE_IN_BYTES
_RESPONSE_STATUS_OFFSET = RESPONSE_HEADER_SIZE
_RESPONSE_MEMBER_UUID_OFFSET = _RESPONSE_STATUS_OFFSET + BYTE_SIZE_IN_BYTES
_RESPONSE_SERIALIZATION_VERSION_OFFSET = _RESPONSE_MEMBER_UUID_OFFSET + UUID_SIZE_IN_BYTES
_RESPONSE_PARTITION_COUNT_OFFSET = _RESPONSE_SERIALIZATION_VERSION_OFFSET + BYTE_SIZE_IN_BYTES
_RESPONSE_CLUSTER_ID_OFFSET = _RESPONSE_PARTITION_COUNT_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_FAILOVER_SUPPORTED_OFFSET = _RESPONSE_CLUSTER_ID_OFFSET + UUID_SIZE_IN_BYTES


def encode_request(cluster_name, credentials, uuid, client_type, serialization_version, client_hazelcast_version, client_name, labels):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_UUID_OFFSET, uuid)
    FixSizedTypesCodec.encode_byte(buf, _REQUEST_SERIALIZATION_VERSION_OFFSET, serialization_version)
    StringCodec.encode(buf, cluster_name)
    ByteArrayCodec.encode(buf, credentials)
    StringCodec.encode(buf, client_type)
    StringCodec.encode(buf, client_hazelcast_version)
    StringCodec.encode(buf, client_name)
    ListMultiFrameCodec.encode(buf, labels, StringCodec.encode, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["status"] = FixSizedTypesCodec.decode_byte(initial_frame.buf, _RESPONSE_STATUS_OFFSET)
    response["member_uuid"] = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_MEMBER_UUID_OFFSET)
    response["serialization_version"] = FixSizedTypesCodec.decode_byte(initial_frame.buf, _RESPONSE_SERIALIZATION_VERSION_OFFSET)
    response["partition_count"] = FixSizedTypesCodec.decode_int(initial_frame.buf, _RESPONSE_PARTITION_COUNT_OFFSET)
    response["cluster_id"] = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_CLUSTER_ID_OFFSET)
    response["failover_supported"] = FixSizedTypesCodec.decode_boolean(initial_frame.buf, _RESPONSE_FAILOVER_SUPPORTED_OFFSET)
    response["address"] = CodecUtil.decode_nullable(msg, AddressCodec.decode)
    response["server_hazelcast_version"] = StringCodec.decode(msg)
    return response



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x030700
_REQUEST_MESSAGE_TYPE = 198400
# hex: 0x030701
_RESPONSE_MESSAGE_TYPE = 198401

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    return CodecUtil.decode_nullable(msg, DataCodec.decode)



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.codec.custom.index_config_codec import IndexConfigCodec

# hex: 0x012900
_REQUEST_MESSAGE_TYPE = 76032
# hex: 0x012901
_RESPONSE_MESSAGE_TYPE = 76033

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name, index_config):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name)
    IndexConfigCodec.encode(buf, index_config, True)
    return OutboundMessage(buf, False)



from hazelcast.future import make_blocking
from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import cp_group_destroy_cp_object_codec


def _no_op_response_handler(_):
    return None


class BaseCPProxy(object):
    def __init__(self, context, group_id, service_name, proxy_name, object_name):
        self._context = context
        self._group_id = group_id
        self._service_name = service_name
        self._proxy_name = proxy_name
        self._object_name = object_name
        self._invocation_service = context.invocation_service
        serialization_service = context.serialization_service
        self._to_data = serialization_service.to_data
        self._to_object = serialization_service.to_object

    def destroy(self):
        """Destroys this proxy."""
        codec = cp_group_destroy_cp_object_codec
        request = codec.encode_request(self._group_id, self._service_name, self._object_name)
        return self._invoke(request)

    def blocking(self):
        """Returns a version of this proxy with only blocking method calls."""
        return make_blocking(self)

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(request, response_handler=response_handler)
        self._invocation_service.invoke(invocation)
        return invocation.future


class SessionAwareCPProxy(BaseCPProxy):
    def __init__(self, context, group_id, service_name, proxy_name, object_name):
        super(SessionAwareCPProxy, self).__init__(
            context, group_id, service_name, proxy_name, object_name
        )
        self._session_manager = context.proxy_session_manager

    def get_group_id(self):
        """
        Returns:
           hazelcast.protocol.RaftGroupId: Id of the CP group that runs this proxy.
        """
        return self._group_id

    def _get_session_id(self):
        """
        Returns:
            int: Session id.
        """
        return self._session_manager.get_session_id(self._group_id)

    def _acquire_session(self, count=1):
        """
        Returns:
            hazelcast.future.Future[int]: Session id.
        """
        return self._session_manager.acquire_session(self._group_id, count)

    def _release_session(self, session_id, count=1):
        self._session_manager.release_session(self._group_id, session_id, count)

    def _invalidate_session(self, session_id):
        self._session_manager.invalidate_session(self._group_id, session_id)



from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, BEGIN_FRAME_BUF
from hazelcast.protocol.builtin import StringCodec
from hazelcast.core import DistributedObjectInfo


class DistributedObjectInfoCodec(object):
    @staticmethod
    def encode(buf, distributed_object_info, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        StringCodec.encode(buf, distributed_object_info.service_name)
        StringCodec.encode(buf, distributed_object_info.name)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        service_name = StringCodec.decode(msg)
        name = StringCodec.decode(msg)
        CodecUtil.fast_forward_to_end_frame(msg)
        return DistributedObjectInfo(service_name, name)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec

# hex: 0x012000
_REQUEST_MESSAGE_TYPE = 73728
# hex: 0x012001
_RESPONSE_MESSAGE_TYPE = 73729

_REQUEST_REPLACE_EXISTING_VALUES_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_REPLACE_EXISTING_VALUES_OFFSET + BOOLEAN_SIZE_IN_BYTES


def encode_request(name, replace_existing_values):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_REPLACE_EXISTING_VALUES_OFFSET, replace_existing_values)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, False)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x010D00
_REQUEST_MESSAGE_TYPE = 68864
# hex: 0x010D01
_RESPONSE_MESSAGE_TYPE = 68865

_REQUEST_THREAD_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_TTL_OFFSET = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TTL_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, key, value, thread_id, ttl):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TTL_OFFSET, ttl)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, key)
    DataCodec.encode(buf, value, True)
    return OutboundMessage(buf, False)
